import inspect
import sys
import warnings
from collections.abc import Collection, Hashable, Iterable, Sequence
from dataclasses import dataclass, fields
from functools import total_ordering
from pathlib import Path
from types import UnionType
from typing import Any, ClassVar, Literal, TypeAlias, TypeVar, get_args, get_origin

import pandas as pd
from cognite.client.data_classes.data_modeling import (
    ContainerId,
    DataModelId,
    PropertyId,
    ViewId,
)

from cognite.neat._utils.text import humanize_collection, to_camel_case, to_snake_case

if sys.version_info < (3, 11):
    from exceptiongroup import ExceptionGroup
    from typing_extensions import Self
else:
    from typing import Self


__all__ = [
    "IssueList",
    "MultiValueError",
    "NeatError",
    "NeatIssue",
    "NeatWarning",
]

T_Identifier = TypeVar("T_Identifier", bound=Hashable)

T_ReferenceIdentifier = TypeVar("T_ReferenceIdentifier", bound=Hashable)

ResourceType: TypeAlias = (
    Literal[
        "view",
        "container",
        "view property",
        "container property",
        "space",
        "class",
        "property",
        "asset",
        "relationship",
        "data model",
        "edge",
        "node",
        "enum collection",
        "unknown",
    ]
    # String to handle all unknown types in different importers.
    | str
)


@total_ordering
@dataclass(unsafe_hash=True)
class NeatIssue:
    """This is the base class for all exceptions and warnings (issues) used in Neat."""

    extra: ClassVar[str | None] = None
    fix: ClassVar[str | None] = None

    def as_message(self, include_type: bool = True) -> str:
        """Return a human-readable message for the issue."""
        template = self.__doc__
        if not template:
            return "Missing"
        variables, has_all_optional = self._get_variables()

        msg = template.format(**variables)
        if self.extra and has_all_optional:
            msg += "\n" + self.extra.format(**variables)
        if self.fix:
            msg += f"\nFix: {self.fix.format(**variables)}"
        if include_type:
            name = type(self).__name__
            msg = f"{name}: {msg}"
        return msg

    def _get_variables(self) -> tuple[dict[str, str], bool]:
        variables: dict[str, str] = {}
        has_all_optional = True
        for name, var_ in vars(self).items():
            if var_ is None:
                has_all_optional = False
            elif isinstance(var_, str):
                variables[name] = var_
            elif isinstance(var_, Path):
                variables[name] = var_.as_posix()
            elif isinstance(var_, Collection):
                variables[name] = humanize_collection(var_)
            elif isinstance(var_, NeatError):
                variables[name] = var_.as_message(include_type=False)
            else:
                variables[name] = repr(var_)
        return variables, has_all_optional

    def dump(self) -> dict[str, Any]:
        """Return a dictionary representation of the issue."""
        variables = vars(self)
        output = {
            to_camel_case(key): self._dump_value(value)
            for key, value in variables.items()
            if not (value is None or key.startswith("_"))
        }
        output["NeatIssue"] = type(self).__name__
        return output

    @classmethod
    def _dump_value(cls, value: Any) -> list | int | bool | float | str | dict:
        from cognite.neat._rules.models.entities import Entity

        if isinstance(value, str | int | bool | float):
            return value
        elif isinstance(value, frozenset):
            return [cls._dump_value(item) for item in value]
        elif isinstance(value, Path):
            return value.as_posix()
        elif isinstance(value, tuple):
            return [cls._dump_value(item) for item in value]
        elif isinstance(value, ViewId | ContainerId):
            return value.dump(camel_case=True, include_type=True)
        elif isinstance(value, Entity):
            return value.dump()
        elif isinstance(value, PropertyId):
            return value.dump(camel_case=True)
        elif isinstance(value, DataModelId):
            return value.dump(camel_case=True, include_type=False)
        elif isinstance(value, NeatError):
            return value.dump()
        raise ValueError(f"Unsupported type: {type(value)}")

    @classmethod
    def load(cls, data: dict[str, Any]) -> "NeatIssue":
        """Create an instance of the issue from a dictionary."""
        from cognite.neat._issues.errors import _NEAT_ERRORS_BY_NAME, NeatValueError
        from cognite.neat._issues.warnings import _NEAT_WARNINGS_BY_NAME

        if "NeatIssue" not in data:
            raise NeatValueError("The data does not contain a NeatIssue key.")
        issue_type = data.pop("NeatIssue")
        args = {to_snake_case(key): value for key, value in data.items()}
        if issue_type in _NEAT_ERRORS_BY_NAME:
            return cls._load_values(_NEAT_ERRORS_BY_NAME[issue_type], args)
        elif issue_type in _NEAT_WARNINGS_BY_NAME:
            return cls._load_values(_NEAT_WARNINGS_BY_NAME[issue_type], args)
        else:
            raise NeatValueError(f"Unknown issue type: {issue_type}")

    @classmethod
    def _load_values(cls, neat_issue_cls: "type[NeatIssue]", data: dict[str, Any]) -> "NeatIssue":
        args: dict[str, Any] = {}
        for f in fields(neat_issue_cls):
            if f.name not in data:
                continue
            value = data[f.name]
            args[f.name] = cls._load_value(f.type, value)
        return neat_issue_cls(**args)

    @classmethod
    def _load_value(cls, type_: Any, value: Any) -> Any:
        from cognite.neat._rules.models.entities import Entity

        if isinstance(type_, UnionType) or get_origin(type_) is UnionType:
            args = get_args(type_)
            return cls._load_value(args[0], value)
        elif type_ is frozenset or get_origin(type_) is frozenset:
            subtype = get_args(type_)[0]
            return frozenset(cls._load_value(subtype, item) for item in value)
        elif type_ is Path:
            return Path(value)
        elif type_ is tuple or get_origin(type_) is tuple:
            subtype = get_args(type_)[0]
            return tuple(cls._load_value(subtype, item) for item in value)
        elif type_ is ViewId:
            return ViewId.load(value)
        elif type_ is DataModelId:
            return DataModelId.load(value)
        elif type_ is PropertyId:
            return PropertyId.load(value)
        elif type_ is ContainerId:
            return ContainerId.load(value)
        elif inspect.isclass(type_) and issubclass(type_, Entity):
            return type_.load(value)
        elif type_ is NeatError:
            return cls.load(value)
        return value

    def __lt__(self, other: "NeatIssue") -> bool:
        if not isinstance(other, NeatIssue):
            return NotImplemented
        return (type(self).__name__, self.as_message()) < (type(other).__name__, other.as_message())

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, NeatIssue):
            return NotImplemented
        return (type(self).__name__, self.as_message()) == (type(other).__name__, other.as_message())

    def __str__(self) -> str:
        return self.as_message()


@dataclass(unsafe_hash=True)
class NeatError(NeatIssue, Exception):
    """This is the base class for all exceptions (errors) used in Neat."""

    ...


@dataclass(unsafe_hash=True)
class NeatWarning(NeatIssue, UserWarning):
    """This is the base class for all warnings used in Neat."""

    ...


class MultiValueError(ValueError):
    """This is a container for multiple errors.

    It is used in the pydantic field_validator/model_validator to collect multiple errors, which
    can then be caught in a try-except block and returned as an IssueList.

    """

    def __init__(self, errors: Sequence[NeatIssue]):
        self.errors = IssueList(errors)


class IssueList(list, Sequence[NeatIssue]):
    """This is a generic list of NeatIssues."""

    def __init__(
        self,
        issues: Sequence[NeatIssue] | None = None,
        title: str | None = None,
        action: str | None = None,
        hint: str | None = None,
    ):
        super().__init__(issues or [])
        self.title = title
        self.action = action
        self.hint = hint

    @property
    def errors(self) -> Self:
        """Return all the errors in this list."""
        return type(self)([issue for issue in self if isinstance(issue, NeatError)])  # type: ignore[misc]

    @property
    def has_errors(self) -> bool:
        """Return True if this list contains any errors."""
        return any(isinstance(issue, NeatError) for issue in self)

    @property
    def has_warnings(self) -> bool:
        """Return True if this list contains any warnings."""
        return any(isinstance(issue, NeatWarning) for issue in self)

    @property
    def warnings(self) -> Self:
        """Return all the warnings in this list."""
        return type(self)([issue for issue in self if isinstance(issue, NeatWarning)])  # type: ignore[misc]

    def has_error_type(self, error_type: type[NeatError]) -> bool:
        """Return True if this list contains any errors of the given type."""
        return any(isinstance(issue, error_type) for issue in self)

    def has_warning_type(self, warning_type: type[NeatWarning]) -> bool:
        """Return True if this list contains any warnings of the given type."""
        return any(isinstance(issue, warning_type) for issue in self)

    def as_errors(self, operation: str = "Operation failed") -> ExceptionGroup:
        """Return an ExceptionGroup with all the errors in this list."""
        return ExceptionGroup(
            operation,
            [issue for issue in self if isinstance(issue, NeatError)],
        )

    def trigger_warnings(self) -> None:
        """Trigger all warnings in this list."""
        for warning in self.warnings:
            warnings.warn(warning, stacklevel=2)

    def to_pandas(self) -> pd.DataFrame:
        """Return a pandas DataFrame representation of this list."""
        return pd.DataFrame([issue.dump() for issue in self])

    def as_exception(self) -> MultiValueError:
        """Return a MultiValueError with all the errors in this list."""
        return MultiValueError(self.errors)

    def _repr_html_(self) -> str | None:
        if self.action and not self:
            header = f"Success: {self.action}"
        elif self.action and self.has_errors:
            header = f"Failed: {self.action}"
        elif self.action and self.has_warnings:
            header = f"Succeeded with warnings: {self.action}"
        elif not self:
            header = "No issues found"
        else:
            header = ""

        body = f"<p>{header}</p>"

        if self:
            df = self.to_pandas()
            agg_df = df["NeatIssue"].value_counts().to_frame()
            if len(agg_df) < 10:
                table = agg_df._repr_html_()  # type: ignore[operator]
            else:
                table = agg_df.head()._repr_html_()  # type: ignore[operator]
            body += f"<br />{table}"

            if self.hint:
                body += f"<br />Hint: {self.hint}"
        return body


T_Cls = TypeVar("T_Cls")


def _get_subclasses(cls_: type[T_Cls], include_base: bool = False) -> Iterable[type[T_Cls]]:
    """Get all subclasses of a class."""
    if include_base:
        yield cls_
    for s in cls_.__subclasses__():
        yield s
        yield from _get_subclasses(s, False)



from collections.abc import Iterable
from pathlib import Path
from urllib.parse import quote

from cognite.client import CogniteClient
from cognite.client.data_classes import Label, LabelDefinition, LabelDefinitionList

from ._base import ClassicCDFBaseExtractor, InstanceIdPrefix, T_CogniteResource


class LabelsExtractor(ClassicCDFBaseExtractor[LabelDefinition]):
    """Extract data from Cognite Data Fusions Labels into Neat."""

    _default_rdf_type = "Label"
    _instance_id_prefix = InstanceIdPrefix.label

    @classmethod
    def _from_dataset(
        cls, client: CogniteClient, data_set_external_id: str
    ) -> tuple[int | None, Iterable[LabelDefinition]]:
        items = client.labels(data_set_external_ids=data_set_external_id)
        return None, items

    @classmethod
    def _from_hierarchy(
        cls, client: CogniteClient, root_asset_external_id: str
    ) -> tuple[int | None, Iterable[T_CogniteResource]]:
        raise NotImplementedError("Hierarchy is not supported for labels")

    @classmethod
    def _from_file(cls, file_path: str | Path) -> tuple[int | None, Iterable[LabelDefinition]]:
        labels = LabelDefinitionList.load(Path(file_path).read_text())
        return len(labels), labels

    def _fallback_id(self, item: LabelDefinition) -> str | None:
        if not item.external_id:
            return None
        return self._label_id(item)

    @staticmethod
    def _label_id(label: Label | LabelDefinition | dict) -> str:
        # external_id can create ill-formed URIs, so we create websafe URIs
        # since labels do not have internal ids, we use the external_id as the id
        external_id: str | None = None
        if isinstance(label, dict):
            if "externalId" in label:
                external_id = label["externalId"]
            elif "external_id" in label:
                external_id = label["external_id"]
        else:
            external_id = label.external_id
        if external_id is None:
            raise ValueError("External id must be set of the label")
        return quote(external_id)



import warnings

from cognite.neat._alpha import AlphaFlags
from cognite.neat._issues._base import IssueList
from cognite.neat._rules.models.entities._single_value import ClassEntity, ViewEntity
from cognite.neat._rules.transformers import SubsetDMSRules, SubsetInformationRules

from ._state import SessionState
from .exceptions import NeatSessionError, session_class_wrapper

try:
    from rich import print
except ImportError:
    ...


@session_class_wrapper
class SubsetAPI:
    """
    Subset data model and instances in the session based on the desired subset of concepts.

    """

    def __init__(self, state: SessionState):
        self._state = state

    def data_model(self, concepts: str | list[str]) -> IssueList:
        """Subset the data model to the desired concepts.

        Args:
            concepts: The concepts to subset the data model to.

        Returns:
            IssueList: A list of issues that occurred during the transformation.

        Example:
            Read the CogniteCore data model and reduce the data model to only the 'CogniteAsset' concept.
            ```python
            neat = NeatSession(CogniteClient())

            neat.read.examples.core_data_model()

            neat.subset.data_model("CogniteAsset")
            ```
        """
        if self._state.rule_store.empty:
            raise NeatSessionError("No rules to set the data model ID.")

        warnings.filterwarnings("default")
        AlphaFlags.data_model_subsetting.warn()

        dms = self._state.rule_store.provenance[-1].target_entity.dms
        information = self._state.rule_store.provenance[-1].target_entity.information

        if dms:
            views = {
                ViewEntity(
                    space=dms.metadata.space,
                    externalId=concept,
                    version=dms.metadata.version,
                )
                for concept in concepts
            }

            issues = self._state.rule_transform(SubsetDMSRules(views=views))
            if not issues:
                after = len(self._state.rule_store.last_verified_dms_rules.views)

        elif information:
            classes = {ClassEntity(prefix=information.metadata.space, suffix=concept) for concept in concepts}

            issues = self._state.rule_transform(SubsetInformationRules(classes=classes))
            if not issues:
                after = len(self._state.rule_store.last_verified_information_rules.classes)

        else:
            raise NeatSessionError("Something went terrible wrong. Please contact the neat team.")

        if not issues:
            print(f"Subset to {after} concepts.")

        return issues



import itertools
import json
import typing
from collections.abc import Iterable, Set
from pathlib import Path
from typing import Any

from cognite.client import CogniteClient
from cognite.client.data_classes import Sequence, SequenceFilter
from rdflib import RDF, XSD, Literal, Namespace, URIRef

from cognite.neat._client.data_classes.neat_sequence import NeatSequence, NeatSequenceList
from cognite.neat._shared import Triple

from ._base import DEFAULT_SKIP_METADATA_VALUES, ClassicCDFBaseExtractor, InstanceIdPrefix


class SequencesExtractor(ClassicCDFBaseExtractor[NeatSequence]):
    """Extract data from Cognite Data Fusions Sequences into Neat.

    Args:
        items (Iterable[T_CogniteResource]): An iterable of classic resource.
        namespace (Namespace, optional): The namespace to use. Defaults to DEFAULT_NAMESPACE.
        total (int, optional): The total number of items to load. If passed, you will get a progress bar if rich
            is installed. Defaults to None.
        limit (int, optional): The maximal number of items to load. Defaults to None. This is typically used for
            testing setup of the extractor. For example, if you are extracting 100 000 assets, you might want to
            limit the extraction to 1000 assets to test the setup.
        unpack_metadata (bool, optional): Whether to unpack metadata. Defaults to False, which yields the metadata as
            a JSON string.
        skip_metadata_values (set[str] | frozenset[str] | None, optional): If you are unpacking metadata, then
           values in this set will be skipped.
        camel_case (bool, optional): Whether to use camelCase instead of snake_case for property names.
            Defaults to True.
        as_write (bool, optional): Whether to use the write/request format of the items. Defaults to False.
        unpack_columns (bool, optional): Whether to unpack columns. Defaults to False.
    """

    _default_rdf_type = "Sequence"
    _column_rdf_type = "ColumnClass"
    _instance_id_prefix = InstanceIdPrefix.sequence

    def __init__(
        self,
        items: Iterable[NeatSequence],
        namespace: Namespace | None = None,
        total: int | None = None,
        limit: int | None = None,
        unpack_metadata: bool = True,
        skip_metadata_values: Set[str] | None = DEFAULT_SKIP_METADATA_VALUES,
        camel_case: bool = True,
        as_write: bool = False,
        prefix: str | None = None,
        identifier: typing.Literal["id", "externalId"] = "id",
        unpack_columns: bool = False,
    ):
        super().__init__(
            items,
            namespace,
            total,
            limit,
            unpack_metadata,
            skip_metadata_values,
            camel_case,
            as_write,
            prefix,
            identifier,
        )
        self.unpack_columns = unpack_columns

    @classmethod
    def from_dataset(
        cls,
        client: CogniteClient,
        data_set_external_id: str,
        namespace: Namespace | None = None,
        limit: int | None = None,
        unpack_metadata: bool = True,
        skip_metadata_values: Set[str] | None = DEFAULT_SKIP_METADATA_VALUES,
        camel_case: bool = True,
        as_write: bool = False,
        prefix: str | None = None,
        identifier: typing.Literal["id", "externalId"] = "id",
        unpack_columns: bool = False,
        skip_rows: bool = False,
    ):
        total, items = cls._handle_no_access(lambda: cls._from_dataset(client, data_set_external_id, skip_rows))
        return cls(
            items,
            namespace,
            total,
            limit,
            unpack_metadata,
            skip_metadata_values,
            camel_case,
            as_write,
            prefix,
            identifier,
            unpack_columns,
        )

    @classmethod
    def from_hierarchy(
        cls,
        client: CogniteClient,
        root_asset_external_id: str,
        namespace: Namespace | None = None,
        limit: int | None = None,
        unpack_metadata: bool = True,
        skip_metadata_values: Set[str] | None = DEFAULT_SKIP_METADATA_VALUES,
        camel_case: bool = True,
        as_write: bool = False,
        prefix: str | None = None,
        identifier: typing.Literal["id", "externalId"] = "id",
        unpack_columns: bool = False,
        skip_rows: bool = False,
    ):
        total, items = cls._handle_no_access(lambda: cls._from_hierarchy(client, root_asset_external_id, skip_rows))
        return cls(
            items,
            namespace,
            total,
            limit,
            unpack_metadata,
            skip_metadata_values,
            camel_case,
            as_write,
            prefix,
            identifier,
            unpack_columns,
        )

    @classmethod
    def from_file(
        cls,
        file_path: str | Path,
        namespace: Namespace | None = None,
        limit: int | None = None,
        unpack_metadata: bool = True,
        skip_metadata_values: Set[str] | None = DEFAULT_SKIP_METADATA_VALUES,
        camel_case: bool = True,
        as_write: bool = False,
        prefix: str | None = None,
        identifier: typing.Literal["id", "externalId"] = "id",
        unpack_columns: bool = False,
    ):
        total, items = cls._from_file(file_path)
        return cls(
            items,
            namespace,
            total,
            limit,
            unpack_metadata,
            skip_metadata_values,
            camel_case,
            as_write,
            prefix,
            identifier,
            unpack_columns,
        )

    @classmethod
    def _from_dataset(
        cls, client: CogniteClient, data_set_external_id: str, skip_rows: bool = False
    ) -> tuple[int | None, Iterable[NeatSequence]]:
        total = client.sequences.aggregate_count(
            filter=SequenceFilter(data_set_ids=[{"externalId": data_set_external_id}])
        )
        items = client.sequences(data_set_external_ids=data_set_external_id)
        if skip_rows:
            return total, (NeatSequence.from_cognite_sequence(seq) for seq in items)
        else:
            return total, cls._lookup_rows(items, client)

    @classmethod
    def _from_hierarchy(
        cls, client: CogniteClient, root_asset_external_id: str, skip_rows: bool = False
    ) -> tuple[int | None, Iterable[NeatSequence]]:
        total = client.sequences.aggregate_count(
            filter=SequenceFilter(asset_subtree_ids=[{"externalId": root_asset_external_id}])
        )
        items = client.sequences(asset_subtree_external_ids=[root_asset_external_id])
        if skip_rows:
            return total, (NeatSequence.from_cognite_sequence(seq) for seq in items)
        else:
            return total, cls._lookup_rows(items, client)

    @classmethod
    def _from_file(cls, file_path: str | Path) -> tuple[int | None, Iterable[NeatSequence]]:
        sequences = NeatSequenceList.load(Path(file_path).read_text())
        return len(sequences), sequences

    @classmethod
    def _lookup_rows(cls, sequence_iterable: Iterable[Sequence], client: CogniteClient) -> Iterable[NeatSequence]:
        iterator = iter(sequence_iterable)
        for sequences in iter(lambda: list(itertools.islice(iterator, client.config.max_workers)), []):
            # The PySDK uses max_workers to limit the number of requests made in parallel.
            # We can only get one set of sequence rows per request, so we chunk the sequences up into groups of
            # max_workers and then make a request to get all the rows for those sequences in one go.
            sequence_list = list(sequences)
            row_list = client.sequences.rows.retrieve(id=[seq.id for seq in sequence_list])
            rows_by_sequence_id = {row.id: row.rows for row in row_list}
            for seq in sequence_list:
                yield NeatSequence.from_cognite_sequence(seq, rows_by_sequence_id.get(seq.id))

    def _item2triples_special_cases(self, id_: URIRef, dumped: dict[str, Any]) -> list[Triple]:
        """For sequences, columns and rows are special cases.'"""
        if self.unpack_columns:
            return self._unpack_columns(id_, dumped)
        else:
            return self._default_columns_and_rows(id_, dumped)

    def _default_columns_and_rows(self, id_: URIRef, dumped: dict[str, Any]) -> list[Triple]:
        triples: list[Triple] = []
        if "columns" in dumped:
            columns = dumped.pop("columns")
            triples.extend(
                [
                    (
                        id_,
                        self.namespace.columns,
                        # Rows have a rowNumber, so we introduce colNumber here to be consistent.
                        Literal(json.dumps({"colNumber": no, **col}), datatype=XSD._NS["json"]),
                    )
                    for no, col in enumerate(columns, 1)
                ]
            )
        if "rows" in dumped:
            rows = dumped.pop("rows")
            triples.extend(
                [(id_, self.namespace.rows, Literal(json.dumps(row), datatype=XSD._NS["json"])) for row in rows]
            )
        return triples

    def _unpack_columns(self, id_: URIRef, dumped: dict[str, Any]) -> list[Triple]:
        triples: list[Triple] = []
        columnValueTypes: list[str] = []
        column_order: list[str] = []
        if columns := dumped.pop("columns", None):
            for col in columns:
                external_id = col.pop("externalId")
                column_order.append(external_id)
                value_type = col.pop("valueType")
                columnValueTypes.append(value_type)

                col_id = self.namespace[f"Column_{external_id}"]
                triples.append((id_, self.namespace[external_id], col_id))
                type_ = self.namespace[self._column_rdf_type]
                triples.append((col_id, RDF.type, type_))
                if metadata := col.pop("metadata", None):
                    triples.extend(self._metadata_to_triples(col_id, metadata))
                # Should only be name and description left in col
                for key, value in col.items():
                    if value is None:
                        continue
                    triples.append((col_id, self.namespace[key], Literal(value, datatype=XSD.string)))

            triples.append(
                (id_, self.namespace.columnOrder, Literal(json.dumps(column_order), datatype=XSD._NS["json"]))
            )
            triples.append(
                (id_, self.namespace.columnValueTypes, Literal(json.dumps(columnValueTypes), datatype=XSD._NS["json"]))
            )
        if rows := dumped.pop("rows", None):
            values_by_column: list[list[Any]] = [[] for _ in column_order]
            for row in rows:
                for i, value in enumerate(row["values"]):
                    values_by_column[i].append(value)
            for col_name, values in zip(column_order, values_by_column, strict=False):
                triples.append(
                    (id_, self.namespace[f"{col_name}Values"], Literal(json.dumps(values), datatype=XSD._NS["json"]))
                )

        return triples



from collections.abc import Iterable
from pathlib import Path

from cognite.client import CogniteClient
from cognite.client.data_classes import Event, EventFilter, EventList

from ._base import ClassicCDFBaseExtractor, InstanceIdPrefix


class EventsExtractor(ClassicCDFBaseExtractor[Event]):
    """Extract data from Cognite Data Fusions Events into Neat."""

    _default_rdf_type = "Event"
    _instance_id_prefix = InstanceIdPrefix.event

    @classmethod
    def _from_dataset(cls, client: CogniteClient, data_set_external_id: str) -> tuple[int | None, Iterable[Event]]:
        total = client.events.aggregate_count(filter=EventFilter(data_set_ids=[{"externalId": data_set_external_id}]))
        items = client.events(data_set_external_ids=data_set_external_id)
        return total, items

    @classmethod
    def _from_hierarchy(cls, client: CogniteClient, root_asset_external_id: str) -> tuple[int | None, Iterable[Event]]:
        total = client.events.aggregate_count(
            filter=EventFilter(asset_subtree_ids=[{"externalId": root_asset_external_id}])
        )
        items = client.events(asset_subtree_external_ids=[root_asset_external_id])
        return total, items

    @classmethod
    def _from_file(cls, file_path: str | Path) -> tuple[int | None, Iterable[Event]]:
        assets = EventList.load(Path(file_path).read_text())
        return len(assets), assets



import warnings
from collections.abc import Callable
from typing import Any

from rdflib import URIRef

from cognite.neat._alpha import AlphaFlags
from cognite.neat._graph.transformers import (
    ConnectionToLiteral,
    ConvertLiteral,
    LiteralToEntity,
    RelationshipAsEdgeTransformer,
)
from cognite.neat._graph.transformers._rdfpath import MakeConnectionOnExactMatch
from cognite.neat._issues import IssueList
from cognite.neat._issues.errors import NeatValueError
from cognite.neat._rules.transformers import PrefixEntities, StandardizeNaming
from cognite.neat._rules.transformers._converters import StandardizeSpaceAndVersion
from cognite.neat._utils.text import humanize_collection

from ._state import SessionState
from .exceptions import NeatSessionError, session_class_wrapper


@session_class_wrapper
class PrepareAPI:
    """Apply various operations on the knowledge graph as a necessary preprocessing step before for instance
    inferring a data model or exporting the knowledge graph to a desired destination.
    """

    def __init__(self, state: SessionState, verbose: bool) -> None:
        self._state = state
        self._verbose = verbose
        self.data_model = DataModelPrepareAPI(state, verbose)
        self.instances = InstancePrepareAPI(state, verbose)


@session_class_wrapper
class InstancePrepareAPI:
    """Operations to perform on instances of data in the knowledge graph."""

    def __init__(self, state: SessionState, verbose: bool) -> None:
        self._state = state
        self._verbose = verbose

    def make_connection_on_exact_match(
        self,
        source: tuple[str, str],
        target: tuple[str, str],
        connection: str | None = None,
        limit: int | None = 100,
    ) -> None:
        """Make connection on exact match.

        Args:
            source: The source of the connection. A tuple of (type, property) where
                    where property is the property that should be matched on the source
                    to make the connection with the target.
            target: The target of the connection. A tuple of (type, property) where
                    where property is the property that should be matched on the target
                    to make the connection with the source.

            connection: new property to use for the connection. If None, the connection
                    will be made by lowercasing the target type.
            limit: The maximum number of connections to make. If None, all connections

        !!! note "Make Connection on Exact Match"
            This method will make a connection between the source and target based on the exact match:
            (SourceType)-[sourceProperty]->(sourceValue) == (TargetType)-[targetProperty]->(targetValue)

            The connection will be made by creating a new property on the source type that will contain the
            target value, as follows:
            (SourceType)-[connection]->(TargetType)

        Example:
            Make connection on exact match:
            ```python
            # From an active NeatSession
            neat.read.csv("workitem.Table.csv",
                          type = "Activity",
                          primary_key="sourceId")

            neat.read.csv("assets.Table.csv",
                          type="Asset",
                          primary_key="WMT_TAG_GLOBALID")

            # Here we specify what column from the source table we should use when we link it with a column in the
            # target table. In this case, it is the "workorderItemname" column in the source table
            source = ("Activity", "workorderItemname")

            # Here we give a name to the new property that is created when a match between the source and target is
            # found
            connection = "asset"

            # Here we specify what column from the target table we should use when searching for a match.
            # In this case, it is the "wmtTagName" column in the target table
            target = ("Asset", "wmtTagName")

            neat.prepare.instances.make_connection_on_exact_match(source, target, connection)
            ```
        """
        try:
            subject_type, subject_predicate = self._get_type_and_property_uris(*source)
            object_type, object_predicate = self._get_type_and_property_uris(*target)
        except NeatValueError as e:
            raise NeatSessionError(f"Cannot make connection: {e}") from None

        transformer = MakeConnectionOnExactMatch(
            subject_type,
            subject_predicate,
            object_type,
            object_predicate,
            connection,
            limit,
        )

        self._state.instances.store.transform(transformer)

    def _get_type_and_property_uris(self, type_: str, property_: str) -> tuple[URIRef, URIRef]:
        type_uri = self._state.instances.store.queries.type_uri(type_)
        property_uri = self._state.instances.store.queries.property_uri(property_)

        if not type_uri:
            raise NeatValueError(f"Type {type_} does not exist in the graph.")
        elif len(type_uri) > 1:
            raise NeatValueError(f"{type_} has multiple ids found in the graph: {humanize_collection(type_uri)}.")

        if not property_uri:
            raise NeatValueError(f"Property {property_} does not exist in the graph.")
        elif len(type_uri) > 1:
            raise NeatValueError(
                f"{property_} has multiple ids found in the graph: {humanize_collection(property_uri)}."
            )

        if not self._state.instances.store.queries.type_with_property(type_uri[0], property_uri[0]):
            raise NeatValueError(f"Property {property_} is not defined for type {type_}.")
        return type_uri[0], property_uri[0]

    def relationships_as_edges(self, min_relationship_types: int = 1, limit_per_type: int | None = None) -> None:
        """This assumes that you have read a classic CDF knowledge graph including relationships.

        This method converts relationships into edges in the graph. This is useful as the
        edges will be picked up as part of the schema connected to Assets, Events, Files, Sequences,
        and TimeSeries in the InferenceImporter.

        Args:
            min_relationship_types: The minimum number of relationship types that must exists to convert those
                relationships to edges. For example, if there is only 5 relationships between Assets and TimeSeries,
                and limit is 10, those relationships will not be converted to edges.
            limit_per_type: The number of conversions to perform per relationship type. For example, if there are 10
                relationships between Assets and TimeSeries, and limit_per_type is 1, only 1 of those relationships
                will be converted to an edge. If None, all relationships will be converted.

        """
        transformer = RelationshipAsEdgeTransformer(min_relationship_types, limit_per_type)
        self._state.instances.store.transform(transformer)

    def convert_data_type(self, source: tuple[str, str], *, convert: Callable[[Any], Any] | None = None) -> None:
        """Convert the data type of the given property.

        This is, for example, useful when you have a boolean property that you want to convert to an enum.

        Args:
            source: The source of the conversion. A tuple of (type, property)
                    where property is the property that should be converted.
            convert: The function to use for the conversion. The function should take the value of the property
                    as input and return the converted value. Default to assume you have a string that should be
                    converted to int, float, bool, or datetime.

        Example:
            Convert a boolean property to a string:
            ```python
            neat.prepare.instances.convert_data_type(
                ("TimeSeries", "isString"),
                convert=lambda is_string: "string" if is_string else "numeric"
            )
            ```

        """
        try:
            subject_type, subject_predicate = self._get_type_and_property_uris(*source)
        except NeatValueError as e:
            raise NeatSessionError(f"Cannot convert data type: {e}") from None

        transformer = ConvertLiteral(subject_type, subject_predicate, convert)
        self._state.instances.store.transform(transformer)

    def property_to_type(self, source: tuple[str | None, str], type: str, new_property: str | None = None) -> None:
        """Convert a property to a new type.

        Args:
            source: The source of the conversion. A tuple of (type, property)
                    where property is the property that should be converted.
                    You can pass (None, property) to covert all properties with the given name.
            type: The new type of the property.
            new_property: Add the identifier as a new property. If None, the new entity will not have a property.

        Example:
            Convert the property 'source' to SourceSystem
            ```python
            neat.prepare.instances.property_to_type(
                (None, "source"), "SourceSystem"
            )
            ```
        """
        subject_type: URIRef | None = None
        if source[0] is not None:
            try:
                subject_type, subject_predicate = self._get_type_and_property_uris(*source)  # type: ignore[arg-type, assignment]
            except NeatValueError as e:
                raise NeatSessionError(f"Cannot convert to type: {e}") from None
        else:
            subject_predicate = self._state.instances.store.queries.property_uri(source[1])[0]

        transformer = LiteralToEntity(subject_type, subject_predicate, type, new_property)
        self._state.instances.store.transform(transformer)

    def connection_to_data_type(self, source: tuple[str | None, str]) -> None:
        """Converts a connection to a data type.

        Args:
            source: The source of the conversion. A tuple of (type, property)
                    where property is the property that should be converted.
                    You can pass (None, property) to covert all properties with the given name.

        Example:

            Convert all properties 'labels' from a connection to a string:

            ```python
            neat.prepare.instances.connection_to_data_type(
                (None, "labels")
            )
            ```

        """
        subject_type: URIRef | None = None
        if source[0] is not None:
            try:
                subject_type, subject_predicate = self._get_type_and_property_uris(*source)  # type: ignore[arg-type, assignment]
            except NeatValueError as e:
                raise NeatSessionError(f"Cannot convert to data type: {e}") from None
        else:
            subject_predicate = self._state.instances.store.queries.property_uri(source[1])[0]
        transformer = ConnectionToLiteral(subject_type, subject_predicate)
        self._state.instances.store.transform(transformer)


@session_class_wrapper
class DataModelPrepareAPI:
    """Operations to perform on a data model as part of a workflow before writing the data model
    to a desired destination.
    """

    def __init__(self, state: SessionState, verbose: bool) -> None:
        self._state = state
        self._verbose = verbose

    def prefix(self, prefix: str) -> IssueList:
        """Prefix all views in the data model with the given prefix.

        Args:
            prefix: The prefix to add to the views in the data model.

        """

        return self._state.rule_transform(PrefixEntities(prefix))  # type: ignore[arg-type]

    def standardize_naming(self) -> IssueList:
        """Standardize the naming of all views/classes/properties in the data model.

        For classes/views/containers, the naming will be standardized to PascalCase.
        For properties, the naming will be standardized to camelCase.
        """
        warnings.filterwarnings("default")
        AlphaFlags.standardize_naming.warn()
        return self._state.rule_transform(StandardizeNaming())

    def standardize_space_and_version(self) -> IssueList:
        """Standardize space and version in the data model.

        This method will standardize the space and version in the data model to the Cognite standard.
        """
        warnings.filterwarnings("default")
        AlphaFlags.standardize_space_and_version.warn()
        return self._state.rule_transform(StandardizeSpaceAndVersion())



import urllib.parse
from collections import defaultdict
from collections.abc import Iterable
from typing import Any, Literal, cast, overload

from rdflib import RDF, XSD, Dataset, Graph, Namespace, URIRef
from rdflib import Literal as RdfLiteral
from rdflib.graph import DATASET_DEFAULT_GRAPH_ID
from rdflib.query import ResultRow

from cognite.neat._constants import NEAT
from cognite.neat._shared import InstanceType
from cognite.neat._utils.rdf_ import remove_instance_ids_in_batch, remove_namespace_from_uri


class Queries:
    """Helper class for storing standard queries for the graph store."""

    def __init__(
        self,
        dataset: Dataset,
        default_named_graph: URIRef | None = None,
    ):
        self.dataset = dataset
        self.default_named_graph = default_named_graph or DATASET_DEFAULT_GRAPH_ID

    def graph(self, named_graph: URIRef | None = None) -> Graph:
        """Get named graph from the dataset to query over"""
        return self.dataset.graph(named_graph or self.default_named_graph)

    def summarize_instances(self, named_graph: URIRef | None = None) -> list[tuple[str, int]]:
        """Summarize instances in the graph store by class and count"""
        query_statement = """ SELECT ?class (COUNT(?instance) AS ?instanceCount)
                             WHERE {
                             ?instance a ?class .
                             }
                             GROUP BY ?class
                             ORDER BY DESC(?instanceCount) """

        return [  # type: ignore[misc]
            (
                remove_namespace_from_uri(cast(URIRef, class_)),
                cast(RdfLiteral, count).value,
            )
            for class_, count in self.graph(named_graph=named_graph).query(query_statement)
        ]

    def types(self, named_graph: URIRef | None = None) -> dict[URIRef, str]:
        """Types and their short form in the graph"""
        query = """SELECT DISTINCT ?type
                   WHERE {?s a ?type .}"""

        return {  # type: ignore[misc, index, arg-type]
            cast(URIRef, type_): remove_namespace_from_uri(cast(URIRef, type_))
            for (type_,) in list(self.graph(named_graph).query(query))
        }

    def type_uri(self, type_: str, named_graph: URIRef | None = None) -> list[URIRef]:
        """Get the URIRef of a type"""
        return [k for k, v in self.types(named_graph).items() if v == type_]

    def properties(self, named_graph: URIRef | None = None) -> dict[URIRef, str]:
        """Properties and their short form in the graph

        Args:
            named_graph: Named graph to query over, default None (default graph)

        """
        query = """SELECT DISTINCT ?property
               WHERE {?s ?property ?o . FILTER(?property != rdf:type)}"""
        return {  # type: ignore[misc, index, arg-type]
            cast(URIRef, type_): remove_namespace_from_uri(cast(URIRef, type_))
            for (type_,) in list(self.graph(named_graph).query(query))
        }

    def properties_by_type(self, named_graph: URIRef | None = None) -> dict[URIRef, dict[URIRef, str]]:
        """Properties and their short form in the graph by type

        Args:
            named_graph: Named graph to query over, default None (default graph)

        """
        query = """SELECT DISTINCT ?type ?property
               WHERE {?s a ?type . ?s ?property ?o . FILTER(?property != rdf:type)}"""
        properties_by_type: dict[URIRef, dict[URIRef, str]] = defaultdict(dict)
        for type_, property_ in cast(ResultRow, list(self.graph(named_graph).query(query))):
            properties_by_type[type_][property_] = remove_namespace_from_uri(property_)  # type: ignore[index]
        return properties_by_type

    def property_uri(self, property_: str, named_graph: URIRef | None = None) -> list[URIRef]:
        """Get the URIRef of a property

        Args:
            property_: Property to find URIRef for
            named_graph: Named graph to query over, default None (default graph)
        """
        return [k for k, v in self.properties(named_graph).items() if v == property_]

    @overload
    def list_instances_ids(
        self, class_uri: None = None, limit: int = -1, named_graph: URIRef | None = None
    ) -> Iterable[tuple[URIRef, URIRef]]: ...

    @overload
    def list_instances_ids(
        self, class_uri: URIRef, limit: int = -1, named_graph: URIRef | None = None
    ) -> Iterable[URIRef]: ...

    def list_instances_ids(
        self, class_uri: URIRef | None = None, limit: int = -1, named_graph: URIRef | None = None
    ) -> Iterable[URIRef] | Iterable[tuple[URIRef, URIRef]]:
        """List all instance IDs

        Args:
            class_uri: Class for which instances are to be found, default None (all instances)
            limit: Max number of instances to return, by default -1 meaning all instances
            named_graph: Named graph to query over, default None (default graph)

        Returns:
            List of class instance URIs
        """
        query = "SELECT DISTINCT ?subject"
        if class_uri:
            query += f" WHERE {{ ?subject a <{class_uri}> .}}"
        else:
            query += " ?type WHERE {{ ?subject a ?type .}}"
        if limit != -1:
            query += f" LIMIT {limit}"
        # MyPy is not very happy with RDFLib, so just ignore the type hinting here
        return (tuple(res) if class_uri is None else res[0] for res in self.graph(named_graph).query(query))  # type: ignore[index, return-value, arg-type]

    def type_with_property(self, type_: URIRef, property_uri: URIRef, named_graph: URIRef | None = None) -> bool:
        """Check if a property exists in the graph store

        Args:
            type_: Type URI to check
            property_uri: Property URI to check
            named_graph: Named graph to query over, default None (default graph)

        Returns:
            True if property exists, False otherwise
        """
        query = f"SELECT ?o WHERE {{ ?s a <{type_}> ; <{property_uri}> ?o .}} Limit 1"
        return bool(list(self.graph(named_graph).query(query)))

    def has_namespace(self, namespace: Namespace, named_graph: URIRef | None = None) -> bool:
        """Check if a namespace exists in the graph store

        Args:
            namespace: Namespace to check
            named_graph: Named graph to query over, default None (default graph)

        Returns:
            True if namespace exists, False otherwise
        """
        query = f"ASK WHERE {{ ?s ?p ?o . FILTER(STRSTARTS(STR(?p), STR(<{namespace}>))) }}"
        return bool(self.graph(named_graph).query(query))

    def has_data(self) -> bool:
        """Check if the graph store has data"""
        return cast(bool, next(iter(self.dataset.query("ASK WHERE { ?s ?p ?o }"))))

    def has_type(self, type_: URIRef, named_graph: URIRef | None = None) -> bool:
        """Check if a type exists in the graph store

        Args:
            type_: Type to check
            named_graph: Named graph to query over, default None (default graph)

        Returns:
            True if type exists, False otherwise
        """
        query = f"ASK WHERE {{ ?s a <{type_}> }}"
        return bool(self.graph(named_graph).query(query))

    def describe(
        self,
        instance_id: URIRef,
        instance_type: URIRef | None = None,
        property_renaming_config: dict | None = None,
        named_graph: URIRef | None = None,
        remove_uri_namespace: bool = True,
    ) -> tuple[URIRef, dict[str | InstanceType, list[Any]]] | None:
        """DESCRIBE instance for a given class from the graph store

        Args:
            instance_id: Instance id for which we want to generate query
            instance_type: Type of the instance, default None (will be inferred from triples)
            property_renaming_config: Dictionary to rename properties, default None (no renaming)
            named_graph: Named graph to query over, default None (default graph)
            remove_uri_namespace: Whether to remove the namespace from the URI, by default True


        Returns:
            Dictionary of instance properties
        """
        property_values: dict[str, list[str] | list[URIRef]] = defaultdict(list)
        for _, predicate, object_ in cast(list[ResultRow], self.graph(named_graph).query(f"DESCRIBE <{instance_id}>")):
            if object_.lower() in [
                "",
                "none",
                "nan",
                "null",
            ]:
                continue

            # set property
            if property_renaming_config and predicate != RDF.type:
                property_ = remove_namespace_from_uri(predicate, validation="prefix")
                renamed_property_ = property_renaming_config.get(predicate, property_)

            elif not property_renaming_config and predicate != RDF.type:
                property_ = remove_namespace_from_uri(predicate, validation="prefix")
                renamed_property_ = property_

            else:
                property_ = RDF.type
                renamed_property_ = property_

            value: Any
            if isinstance(object_, URIRef) and remove_uri_namespace:
                # These properties contain the space in the Namespace.
                value = remove_namespace_from_uri(object_, validation="prefix")
            elif isinstance(object_, URIRef):
                value = object_
            elif isinstance(object_, RdfLiteral):
                if object_.datatype == XSD._NS["json"]:
                    # For JSON literals, the .toPython() returns a Literal object.
                    value = str(object_)
                else:
                    value = object_.toPython()
            else:
                # It is a blank node
                value = str(object_)

            # add type to the dictionary
            if predicate != RDF.type:
                property_values[renamed_property_].append(value)  # type: ignore[arg-type]
            else:
                # guarding against multiple rdf:type values as this is not allowed in CDF
                if RDF.type not in property_values:
                    property_values[RDF.type].append(
                        remove_namespace_from_uri(instance_type, validation="prefix") if instance_type else value  # type: ignore[arg-type]
                    )
                else:
                    # we should not have multiple rdf:type values
                    continue
        if property_values:
            return (
                instance_id,
                property_values,
            )
        else:
            return None

    def list_triples(self, limit: int = 25, named_graph: URIRef | None = None) -> list[ResultRow]:
        """List triples in the graph store

        Args:
            limit: Max number of triples to return, by default 25
            named_graph: Named graph to query over, default None (default graph)

        Returns:
            List of triples
        """
        query = f"SELECT ?subject ?predicate ?object WHERE {{ ?subject ?predicate ?object }} LIMIT {limit}"
        return cast(list[ResultRow], list(self.graph(named_graph).query(query)))

    @overload
    def list_types(self, remove_namespace: Literal[False] = False, limit: int = 25) -> list[ResultRow]: ...

    @overload
    def list_types(
        self,
        remove_namespace: Literal[True],
        limit: int = 25,
        named_graph: URIRef | None = None,
    ) -> list[str]: ...

    def list_types(
        self,
        remove_namespace: bool = False,
        limit: int | None = 25,
        named_graph: URIRef | None = None,
    ) -> list[ResultRow] | list[str]:
        """List types in the graph store

        Args:
            limit: Max number of types to return, by default 25
            remove_namespace: Whether to remove the namespace from the type, by default False

        Returns:
            List of types
        """
        query = "SELECT DISTINCT ?type WHERE { ?subject a ?type }"
        if limit is not None:
            query += f" LIMIT {limit}"
        result = cast(list[ResultRow], list(self.graph(named_graph).query(query)))
        if remove_namespace:
            return [remove_namespace_from_uri(res[0]) for res in result]
        return result

    def multi_value_type_property(
        self,
        named_graph: URIRef | None = None,
    ) -> Iterable[tuple[URIRef, URIRef, list[URIRef]]]:
        query = """SELECT ?sourceType ?property
                          (GROUP_CONCAT(DISTINCT STR(?valueType); SEPARATOR=",") AS ?valueTypes)

                   WHERE {{
                       ?s ?property ?o .
                       ?s a ?sourceType .
                       OPTIONAL {{ ?o a ?type }}

                       # Key part to determine value type: either object, data or unknown
                       BIND(   IF(isLiteral(?o),DATATYPE(?o),
                               IF(BOUND(?type), ?type,
                                               <{unknownType}>)) AS ?valueType)
                   }}

                   GROUP BY ?sourceType ?property
                   HAVING (COUNT(DISTINCT ?valueType) > 1)"""

        for (
            source_type,
            property_,
            value_types,
        ) in cast(
            ResultRow,
            self.graph(named_graph).query(query.format(unknownType=str(NEAT.UnknownType))),
        ):
            yield cast(URIRef, source_type), cast(URIRef, property_), [URIRef(uri) for uri in value_types.split(",")]

    def drop_types(
        self,
        type_: list[URIRef],
        named_graph: URIRef | None = None,
    ) -> dict[URIRef, int]:
        """Drop types from the graph store

        Args:
            type_: List of types to drop
            named_graph: Named graph to query over, default None (default graph

        Returns:
            Dictionary of dropped types
        """
        dropped_types: dict[URIRef, int] = {}
        for t in type_:
            instance_ids = list(self.list_instances_ids(t))
            dropped_types[t] = len(instance_ids)
            remove_instance_ids_in_batch(self.graph(named_graph), instance_ids)
        return dropped_types

    def multi_type_instances(self, named_graph: URIRef | None = None) -> dict[str, list[str]]:
        """Find instances with multiple types

        Args:
            named_graph: Named graph to query over, default None (default graph)

        """

        query = """
        SELECT ?instance (GROUP_CONCAT(str(?type); SEPARATOR=",") AS ?types)
        WHERE {
            ?instance a ?type .
        }
        GROUP BY ?instance
        HAVING (COUNT(?type) > 1)
        """

        result = {}
        for instance, types in self.graph(named_graph).query(query):  # type: ignore
            result[remove_namespace_from_uri(instance)] = remove_namespace_from_uri(types.split(","))

        return result

    def count_of_type(self, class_uri: URIRef, named_graph: URIRef | None = None) -> int:
        query = f"SELECT (COUNT(?instance) AS ?instanceCount) WHERE {{ ?instance a <{class_uri}> }}"
        return int(next(iter(self.graph(named_graph).query(query)))[0])  # type: ignore[arg-type, index]

    def types_with_instance_and_property_count(
        self, remove_namespace: bool = True, named_graph: URIRef | None = None
    ) -> list[dict[str, Any]]:
        query = """
        SELECT ?type (COUNT(DISTINCT ?instance) AS ?instanceCount) (COUNT(DISTINCT ?property) AS ?propertyCount)
                WHERE {
                  ?instance a ?type .
                  ?instance ?property ?value .
                  FILTER(?property != rdf:type)
                }
                GROUP BY ?type
                ORDER BY DESC(?instanceCount)"""
        return [
            {
                "type": urllib.parse.unquote(remove_namespace_from_uri(type_)) if remove_namespace else type_,
                "instanceCount": cast(RdfLiteral, instance_count).toPython(),
                "propertyCount": cast(RdfLiteral, property_count).toPython(),
            }
            for type_, instance_count, property_count in list(
                cast(list[ResultRow], self.graph(named_graph).query(query))
            )
        ]

    def properties_with_count(
        self, remove_namespace: bool = True, named_graph: URIRef | None = None
    ) -> list[dict[str, Any]]:
        instance_count_by_type = {
            entry["type"]: entry["instanceCount"]
            for entry in self.types_with_instance_and_property_count(remove_namespace=False, named_graph=named_graph)
        }
        query = """SELECT ?type ?property (COUNT(DISTINCT ?instance) AS ?instanceCount)
WHERE {
  ?instance a ?type .
  ?instance ?property ?value .
  FILTER(?property != rdf:type)
}
GROUP BY ?type ?property
ORDER BY ASC(?type) ASC(?property)"""
        return [
            {
                "type": urllib.parse.unquote(remove_namespace_from_uri(type_)) if remove_namespace else type_,
                "property": urllib.parse.unquote(remove_namespace_from_uri(property)) if remove_namespace else property,
                "instanceCount": cast(RdfLiteral, instance_count).toPython(),
                "total": instance_count_by_type[type_],
            }
            for type_, property, instance_count in list(cast(list[ResultRow], self.graph(named_graph).query(query)))
        ]

    @overload
    def instances_with_properties(
        self, type: URIRef, remove_namespace: Literal[False], named_graph: URIRef | None = None
    ) -> dict[URIRef, set[URIRef]]: ...

    @overload
    def instances_with_properties(
        self, type: URIRef, remove_namespace: Literal[True], named_graph: URIRef | None = None
    ) -> dict[str, set[str]]: ...

    def instances_with_properties(
        self, type: URIRef, remove_namespace: bool = True, named_graph: URIRef | None = None
    ) -> dict[str, set[str]] | dict[URIRef, set[URIRef]]:
        query = """SELECT DISTINCT ?instance ?property
WHERE {{
    ?instance a <{type}> .
    ?instance ?property ?value .
    FILTER(?property != rdf:type)
}}"""
        result = defaultdict(set)
        for instance, property_ in cast(Iterable[ResultRow], self.graph(named_graph).query(query.format(type=type))):
            instance_str = urllib.parse.unquote(remove_namespace_from_uri(instance)) if remove_namespace else instance
            property_str = urllib.parse.unquote(remove_namespace_from_uri(property_)) if remove_namespace else property_
            result[instance_str].add(property_str)
        return result

    def list_instances_ids_by_space(
        self, space_property: URIRef, named_graph: URIRef | None = None
    ) -> Iterable[tuple[URIRef, str]]:
        """Returns instance ids by space"""
        query = f"""SELECT DISTINCT ?instance ?space
                   WHERE {{?instance <{space_property}> ?space}}"""

        for result in cast(Iterable[ResultRow], self.graph(named_graph).query(query)):
            instance_id, space = cast(tuple[URIRef, URIRef | RdfLiteral], result)
            if isinstance(space, URIRef):
                yield instance_id, remove_namespace_from_uri(space)
            elif isinstance(space, RdfLiteral):
                yield instance_id, str(space.toPython())
            else:
                yield instance_id, str(space)



from typing import cast

from rdflib import Graph, Namespace, URIRef
from rdflib.query import ResultRow

from cognite.neat._constants import DEFAULT_NAMESPACE
from cognite.neat._shared import Triple
from cognite.neat._utils.rdf_ import as_neat_compliant_uri
from cognite.neat._utils.text import sentence_or_string_to_camel

from ._base import BaseTransformer, BaseTransformerStandardised, RowTransformationOutput


# TODO: Standardise after figuring out the bug which appears when running test_iodd_transformers.py
class AttachPropertyFromTargetToSource(BaseTransformer):
    """
    Transformer that considers a TargetNode and SourceNode relationship, to extract a property that is attached to
    the TargetNode, and attaches it to the SourceNode instead, while also deleting the edge between
    the SourceNode and TargetNode.
    This means that you no longer have to go via the SourceNode to TargetNode to extract
    the desired property from TargetNode, you can get it directly from the SourceNode instead.
    Further, there are two ways of defining the predicate for the new property to attach to
    the SourceNode. The predicate that is used will either be the old predicate between the SourceNode and TargetNode,
    or, the TargetNode may hold a property with a value for the new predicate to use.
    In this case, the user must specify the name of this predicate property connected to the TargetNode.
    Consider the following example for illustration:

        Ex. AttachPropertyFromTargetToSource
        Graph before transformation:

            :SourceNode a :SourceType .
            :SourceNode :sourceProperty :TargetNode .

            :TargetNode a :TargetType .
            :TargetNode :propertyWhichValueWeWant 'Target Value' .
            :TargetNode :propertyWhichValueWeMightWantAsNameForNewProperty 'PropertyName'

        Use case A after transformation - attach new property to SourceNode using old predicate:

            :SourceNode a :SourceType .
            :SourceNode :sourceProperty 'Target Value' .

        Use case B after transformation - extract new predicate from one of the properties of the TargetNode:

            :SourceNode a :SourceType .
            :SourceNode :PropertyName 'Target Value' .


    The user can provide a flag to decide if the intermediate target node should be removed from the graph or not
    after connecting the target property to the source node. The example illustrates this.
    The default however is False.

    If delete_target_node is not set, the expected number of triples after this transformation should be the same as
    before the transformation.

    If delete_target_node is set, the expected number of triples should be:
        #triples_before - #target_nodes * #target_nodes_properties

        Number of triples after operation from above example: 5 - 1*3 = 2

    Args:
        target_node_type: RDF.type of edge Node
        target_property: URIRef of the property that holds the value attached to the intermediate node
        target_property_holding_new_property: URIRef of the property which value will be new
        property that will be added to the source node
        delete_target_node: bool if the intermediate Node and Edge between source Node
                                and target property should be deleted. Defaults to False.
        convert_literal_to_uri: bool if the value of the new property should be converted to URIRef. Defaults to False.
        namespace: Namespace to use when converting value to URIRef. Defaults to DEFAULT_NAMESPACE.
    """

    description: str = "Attaches a target property from a target node that is connected to a source node."

    _query_template_use_case_a: str = """
    SELECT ?sourceNode ?sourceProperty ?targetNode ?newSourceProperty ?newSourcePropertyValue WHERE {{
        ?sourceNode ?sourceProperty ?targetNode .
        BIND( <{target_property}> as ?newSourceProperty ) .
        ?targetNode a <{target_node_type}> .
        ?targetNode <{target_property}> ?newSourcePropertyValue . }}"""

    _query_template_use_case_b: str = """
    SELECT ?sourceNode ?sourceProperty ?targetNode ?newSourceProperty ?newSourcePropertyValue WHERE {{
        ?sourceNode ?sourceProperty ?targetNode .
        ?targetNode a <{target_node_type}> .
        ?targetNode <{target_property_holding_new_property_name}> ?newSourceProperty .
        ?targetNode <{target_property}> ?newSourcePropertyValue . }}"""

    def __init__(
        self,
        target_node_type: URIRef,
        target_property: URIRef,
        target_property_holding_new_property: URIRef | None = None,
        delete_target_node: bool = False,
        convert_literal_to_uri: bool = False,
        namespace: Namespace | None = None,
    ):
        self.target_node_type = target_node_type
        self.target_property = target_property
        self.delete_target_node = delete_target_node
        self.target_property_holding_new_property = target_property_holding_new_property
        self.convert_literal_to_uri = convert_literal_to_uri
        self.namespace = namespace or DEFAULT_NAMESPACE

    def transform(self, graph) -> None:
        nodes_to_delete: list[tuple] = []

        if self.target_property_holding_new_property is not None:
            query = self._query_template_use_case_b.format(
                target_node_type=self.target_node_type,
                target_property_holding_new_property_name=self.target_property_holding_new_property,
                target_property=self.target_property,
            )
        else:
            query = self._query_template_use_case_a.format(
                target_node_type=self.target_node_type,
                target_property=self.target_property,
            )

        for (
            source_node,
            old_predicate,
            target_node,
            new_predicate_value,
            new_property_value,
        ) in graph.query(query):
            if self.target_property_holding_new_property is not None:
                # Ensure new predicate is URI compliant as we are creating a new predicate
                new_predicate_value_string = sentence_or_string_to_camel(str(new_predicate_value))
                predicate = as_neat_compliant_uri(self.namespace[new_predicate_value_string])
            else:
                predicate = old_predicate
            # Create new connection from source node to value
            graph.add(
                (
                    source_node,
                    predicate,
                    (self.namespace[new_property_value] if self.convert_literal_to_uri else new_property_value),
                )
            )
            # Remove old relationship between source node and destination node
            graph.remove((source_node, old_predicate, target_node))

            nodes_to_delete.append(target_node)

        if self.delete_target_node:
            for target_node in nodes_to_delete:
                # Remove triples with edges to target_node
                graph.remove((None, None, target_node))
                # Remove target node triple and its properties
                graph.remove((target_node, None, None))


# TODO: Remove or adapt IODD
class PruneDanglingNodes(BaseTransformer):
    """
    Knowledge graph pruner and resolver. Will remove rdf triples from graph that does not have connections
    to other nodes, and traverse graph for specified types to resolve the value in the final node and link it to
    the source node.

        Ex. PruneDanglingNodes

        Graph before pruning:
        node(A, rdf:type(Pump)) -> node(B, rdf:type(Disc)),
        node(C, rdf:type(Disc))

        Graph after pruning of nodes rdf:type(Disc):

        node(A, rd:type(Pump)) -> node(B, rdf:type(Disc))

    Args:
        node_prune_types: list of RDF types to prune from the Graph if they are stand-alone Nodes
    """

    description: str = "Prunes nodes of specific rdf types that do not have any connection to them."
    _query_template = """
                    SELECT ?subject
                    WHERE {{
                        ?subject a <{rdf_type}> .
                        FILTER NOT EXISTS {{ ?s ?p ?subject }}
                    }}
            """

    def __init__(
        self,
        node_prune_types: list[URIRef],
    ):
        self.node_prune_types = node_prune_types

    def transform(self, graph: Graph) -> None:
        for type_ in self.node_prune_types:
            for (subject,) in list(graph.query(self._query_template.format(rdf_type=type_))):  # type: ignore
                graph.remove((subject, None, None))


class PruneTypes(BaseTransformerStandardised):
    """
    Removes all the instances of specific type
    """

    description: str = "Prunes nodes of specific rdf types"

    def __init__(
        self,
        node_prune_types: list[URIRef],
    ):
        self.node_prune_types = node_prune_types

    def _iterate_query(self) -> str:
        filter_string = ""
        for node in self.node_prune_types:
            filter_string += f" <{node}> "

        query = """
                SELECT ?subject
                WHERE {{
                  ?subject a ?type .
                  VALUES ?type {{ {rdf_types_string} }}
                }}
                """
        return query.format(rdf_types_string=filter_string)

    def _count_query(self) -> str:
        filter_string = ""
        for node in self.node_prune_types:
            filter_string += f" <{node}> "

        query = """
                SELECT ( COUNT( ?subject ) as ?count )
                WHERE {{
                  ?subject a ?type .
                  VALUES ?type {{ {rdf_types_string} }}
                }}
                """
        return query.format(rdf_types_string=filter_string)

    def operation(self, query_result_row: ResultRow) -> RowTransformationOutput:
        row_output = RowTransformationOutput()

        (subject,) = query_result_row
        row_output.remove_triples.add((subject, None, None))  # type: ignore
        row_output.instances_removed_count = 1

        return row_output


class PruneDeadEndEdges(BaseTransformerStandardised):
    """
    Removes all the triples where object is a node that is not found in graph
    """

    description: str = "Pruning the graph of triples where object is a node that is not found in graph."

    def _iterate_query(self) -> str:
        return """
                 SELECT ?subject ?predicate ?object
                 WHERE {
                     ?subject ?predicate ?object .
                     FILTER (isIRI(?object) && ?predicate != rdf:type)
                     FILTER NOT EXISTS {?object ?p ?o .}
                     }
                 """

    def _count_query(self) -> str:
        return """
                SELECT (COUNT(?object) AS ?count)
                WHERE {
                    ?subject ?predicate ?object .
                    FILTER (isIRI(?object) && ?predicate != rdf:type)
                    FILTER NOT EXISTS {?object ?p ?o .}
                    }
                """

    def operation(self, row: ResultRow) -> RowTransformationOutput:
        row_output = RowTransformationOutput()
        row_output.remove_triples.add(cast(Triple, row))
        row_output.instances_modified_count = 1

        return row_output


class PruneInstancesOfUnknownType(BaseTransformerStandardised):
    """
    Removes all the triples where object is a node that is not found in graph
    """

    description: str = "Prunes the graph of triples where the object is a node that is not found in the graph."

    def _iterate_query(self) -> str:
        return """
                SELECT DISTINCT ?subject
                WHERE {
                    ?subject ?p ?o .
                    FILTER NOT EXISTS {?subject a ?object .}

                    }
                """

    def _count_query(self) -> str:
        return """
                SELECT (COUNT(DISTINCT ?subject) as ?count)
                WHERE {
                    ?subject ?p ?o .
                    FILTER NOT EXISTS {?subject a ?object .}
                    }
                """

    def operation(self, query_result_row: ResultRow) -> RowTransformationOutput:
        row_output = RowTransformationOutput()
        (subject,) = query_result_row
        row_output.remove_triples.add(cast(Triple, (subject, None, None)))
        row_output.instances_removed_count = 1

        return row_output



import hashlib
from collections import defaultdict
from collections.abc import Callable, Hashable
from dataclasses import dataclass
from datetime import datetime, timezone
from functools import partial
from pathlib import Path
from typing import Any, cast

import rdflib
from rdflib import URIRef

from cognite.neat._client import NeatClient
from cognite.neat._constants import DEFAULT_NAMESPACE
from cognite.neat._graph.extractors import DMSGraphExtractor, KnowledgeGraphExtractor
from cognite.neat._issues import IssueList, catch_issues
from cognite.neat._issues.errors import NeatValueError
from cognite.neat._rules._shared import T_VerifiedRules, VerifiedRules
from cognite.neat._rules.exporters import BaseExporter
from cognite.neat._rules.exporters._base import CDFExporter, T_Export
from cognite.neat._rules.importers import BaseImporter
from cognite.neat._rules.models import DMSRules, InformationRules
from cognite.neat._rules.transformers import DMSToInformation, VerifiedRulesTransformer, VerifyAnyRules
from cognite.neat._utils.upload import UploadResultList

from ._provenance import (
    EXTERNAL_AGENT,
    UNKNOWN_AGENT,
    Activity,
    Change,
    Entity,
    Provenance,
)
from .exceptions import EmptyStore, InvalidActivityInput


@dataclass(frozen=True)
class RulesEntity(Entity):
    information: InformationRules
    dms: DMSRules | None = None

    @property
    def has_dms(self) -> bool:
        return self.dms is not None

    @property
    def display_name(self) -> str:
        if self.dms is not None:
            return self.dms.display_name
        return self.information.display_name


@dataclass(frozen=True)
class OutcomeEntity(Entity):
    result: UploadResultList | Path | str | URIRef


class NeatRulesStore:
    def __init__(self) -> None:
        self.provenance = Provenance[RulesEntity]()
        self.exports_by_source_entity_id: dict[rdflib.URIRef, list[Change[OutcomeEntity]]] = defaultdict(list)
        self._last_outcome: UploadResultList | None = None
        self._iteration_by_id: dict[Hashable, int] = {}
        self._last_issues: IssueList | None = None

    def calculate_provenance_hash(self, shorten: bool = True) -> str:
        sha256_hash = hashlib.sha256()
        for change in self.provenance:
            for id_ in [change.agent.id_, change.activity.id_, change.target_entity.id_, change.source_entity.id_]:
                sha256_hash.update(str(id_).encode("utf-8"))
        calculated_hash = sha256_hash.hexdigest()
        if shorten:
            return calculated_hash[:8]
        return calculated_hash

    def _rules_import_verify_convert(
        self,
        importer: BaseImporter,
        validate: bool,
        client: NeatClient | None = None,
    ) -> tuple[InformationRules, DMSRules | None]:
        """Action that imports rules, verifies them and optionally converts them."""
        read_rules = importer.to_rules()
        verified = VerifyAnyRules(validate, client).transform(read_rules)  # type: ignore[arg-type]
        if isinstance(verified, InformationRules):
            return verified, None
        elif isinstance(verified, DMSRules):
            return DMSToInformation().transform(verified), verified
        else:
            # Bug in the code
            raise ValueError(f"Invalid output from importer: {type(verified)}")

    def _graph_import_verify_convert(
        self,
        extractor: KnowledgeGraphExtractor,
    ) -> tuple[InformationRules, DMSRules | None]:
        info = extractor.get_information_rules()
        dms: DMSRules | None = None
        if isinstance(extractor, DMSGraphExtractor):
            dms = extractor.get_dms_rules()
        return info, dms

    def _manual_transform(
        self, importer: BaseImporter, validate: bool = True, client: NeatClient | None = None
    ) -> IssueList:
        result, issue_list, start, end = self._do_activity(
            partial(self._rules_import_verify_convert, importer, validate, client)
        )

        if not result:
            return issue_list

        info, dms = result
        last_change = self.provenance[-1]

        outside_agent = EXTERNAL_AGENT
        outside_activity = Activity(
            was_associated_with=outside_agent,
            started_at_time=last_change.activity.ended_at_time,
            ended_at_time=end,
            used=last_change.target_entity,
        )

        # Case 1: Source of imported rules is not known
        if not (source_id := self._get_source_id(result)):
            raise NeatValueError(
                "The source of the imported rules is unknown."
                " Import will be skipped. Start a new NEAT session and import the data model there."
            )

        # Case 2: Source of imported rules not in rules_store
        if not (source_entity := self.provenance.target_entity(source_id)) or not isinstance(
            source_entity, RulesEntity
        ):
            raise NeatValueError(
                "The source of the imported rules is not in the provenance."
                " Import will be skipped. Start a new NEAT session and import the data model there."
            )

        # Case 3: Source is not the latest source entity in the provenance change
        if source_entity.id_ != last_change.target_entity.id_:
            raise NeatValueError(
                "Imported rules are detached from the provenance chain."
                " Import will be skipped. Start a new NEAT session and import the data model there."
            )

        # Case 4: Provenance is already at the physical state of the data model, going back to logical not possible
        if not dms and source_entity.dms:
            raise NeatValueError(
                "Rules are already in physical state, import of logical model not possible."
                " Import will be skipped. Start a new NEAT session and import the data model there."
            )

        # modification took place on information rules
        if not dms and not source_entity.dms:
            outside_target_entity = RulesEntity(
                was_attributed_to=outside_agent,
                was_generated_by=outside_activity,
                information=info,
                dms=dms,
                issues=issue_list,
                id_=self._create_id(info, dms),
            )

        # modification took place on dms rules, keep latest information rules
        elif dms and source_entity.dms:
            outside_target_entity = RulesEntity(
                was_attributed_to=outside_agent,
                was_generated_by=outside_activity,
                information=last_change.target_entity.information,
                dms=dms,
                issues=issue_list,
                id_=self._create_id(info, dms),
            )

        else:
            raise NeatValueError("Invalid state of rules for manual transformation")

        outside_change = Change(
            source_entity=last_change.target_entity,
            agent=outside_agent,
            activity=outside_activity,
            target_entity=outside_target_entity,
            description="Manual transformation of rules outside of NEAT",
        )

        self._last_issues = issue_list
        # record change that took place outside of neat
        self.provenance.append(outside_change)

        return issue_list

    def import_graph(self, extractor: KnowledgeGraphExtractor) -> IssueList:
        if not self.empty:
            raise NeatValueError(f"Data model already exists. Cannot import {extractor.source_uri}.")
        else:
            return self.do_activity(partial(self._graph_import_verify_convert, extractor), extractor)

    def import_rules(
        self,
        importer: BaseImporter,
        validate: bool = True,
        client: NeatClient | None = None,
        enable_manual_edit: bool = False,
    ) -> IssueList:
        if self.empty:
            return self.do_activity(
                partial(self._rules_import_verify_convert, importer, validate, client),
                importer,
            )
        elif enable_manual_edit:
            return self._manual_transform(importer, validate, client)
        else:
            raise NeatValueError("Re-importing rules in the rules store is not allowed.")

    def transform(self, *transformer: VerifiedRulesTransformer) -> IssueList:
        if not self.provenance:
            raise EmptyStore()

        all_issues = IssueList()
        for agent_tool in transformer:

            def action(
                transformer_item=agent_tool,
            ) -> tuple[InformationRules, DMSRules | None]:
                last_change = self.provenance[-1]
                source_entity = last_change.target_entity
                transformer_input = self._get_transformer_input(source_entity, transformer_item)
                transformer_output = transformer_item.transform(transformer_input)
                if isinstance(transformer_output, InformationRules):
                    return transformer_output, None
                return last_change.target_entity.information, transformer_output

            issues = self.do_activity(action, agent_tool)
            all_issues.extend(issues)

        return all_issues

    def export(self, exporter: BaseExporter[T_VerifiedRules, T_Export]) -> T_Export:
        return self._export_activity(exporter.export, exporter, DEFAULT_NAMESPACE["export-result"])

    def export_to_file(self, exporter: BaseExporter, path: Path) -> None:
        def export_action(input_: VerifiedRules) -> Path:
            exporter.export_to_file(input_, path)
            return path

        self._export_activity(export_action, exporter, DEFAULT_NAMESPACE[path.name])

    def export_to_cdf(self, exporter: CDFExporter, client: NeatClient, dry_run: bool) -> UploadResultList:
        return self._export_activity(
            exporter.export_to_cdf, exporter, DEFAULT_NAMESPACE["upload-result"], client, dry_run
        )

    def do_activity(
        self,
        action: Callable[[], tuple[InformationRules, DMSRules | None]],
        agent_tool: BaseImporter | VerifiedRulesTransformer | KnowledgeGraphExtractor,
    ):
        result, issue_list, start, end = self._do_activity(action)
        self._last_issues = issue_list

        if result:
            self._update_provenance(agent_tool, result, issue_list, start, end)
        return issue_list

    def _update_provenance(
        self,
        agent_tool: BaseImporter | VerifiedRulesTransformer | KnowledgeGraphExtractor,
        result: tuple[InformationRules, DMSRules | None],
        issue_list: IssueList,
        activity_start: datetime,
        activity_end: datetime,
    ) -> None:
        # set source entity
        if isinstance(agent_tool, BaseImporter | KnowledgeGraphExtractor):
            source_entity = Entity.create_with_defaults(
                was_attributed_to=UNKNOWN_AGENT,
                id_=agent_tool.source_uri,
            )
        else:
            source_entity = self.provenance[-1].target_entity

        # setting the rest of provenance components
        info, dms = result
        agent = agent_tool.agent
        activity = Activity(
            was_associated_with=agent,
            ended_at_time=activity_end,
            started_at_time=activity_start,
            used=source_entity,
        )

        target_entity = RulesEntity(
            was_attributed_to=agent,
            was_generated_by=activity,
            information=info,
            dms=dms,
            issues=issue_list,
            # here id can be bumped in case id already exists
            id_=self._create_id(info, dms),
        )
        change = Change(
            agent=agent,
            activity=activity,
            target_entity=target_entity,
            description=agent_tool.description,
            source_entity=source_entity,
        )

        self.provenance.append(change)

    def _do_activity(
        self,
        action: Callable[[], tuple[InformationRules, DMSRules | None]],
    ):
        """This private method is used to execute an activity and return the result and issues."""
        start = datetime.now(timezone.utc)
        result: tuple[InformationRules, DMSRules | None] | None = None
        with catch_issues() as issue_list:
            result = action()
        end = datetime.now(timezone.utc)
        return result, issue_list, start, end

    def _export_activity(self, action: Callable, exporter: BaseExporter, target_id: URIRef, *exporter_args: Any) -> Any:
        if self.empty:
            raise EmptyStore()
        last_change = self.provenance[-1]
        source_entity = last_change.target_entity
        expected_types = exporter.source_types()

        if source_entity.dms is not None and isinstance(source_entity.dms, expected_types):
            input_ = cast(VerifiedRules, source_entity.dms).model_copy(deep=True)
        elif isinstance(source_entity.information, expected_types):
            input_ = cast(VerifiedRules, source_entity.information).model_copy(deep=True)
        else:
            available: list[type] = [InformationRules]
            if source_entity.dms is not None:
                available.append(DMSRules)
            raise InvalidActivityInput(expected=expected_types, have=tuple(available))

        # need to write source prior the export
        input_.metadata.source_id = source_entity.id_

        agent = exporter.agent
        start = datetime.now(timezone.utc)
        with catch_issues() as issue_list:
            # Validate the type of the result
            result = action(input_, *exporter_args)

        end = datetime.now(timezone.utc)
        self._last_issues = issue_list
        activity = Activity(
            was_associated_with=agent,
            ended_at_time=end,
            started_at_time=start,
            used=source_entity,
        )
        if isinstance(result, UploadResultList | Path | URIRef):
            outcome_result: UploadResultList | Path | URIRef | str = result
        else:
            outcome_result = type(result).__name__

        target_entity = OutcomeEntity(
            was_attributed_to=agent,
            was_generated_by=activity,
            result=outcome_result,
            issues=issue_list,
            id_=target_id,
        )
        change = Change(
            agent=agent,
            activity=activity,
            target_entity=target_entity,
            description=exporter.description,
            source_entity=source_entity,
        )

        self.exports_by_source_entity_id[source_entity.id_].append(change)
        if isinstance(result, UploadResultList):
            self._last_outcome = result
        return result

    @staticmethod
    def _get_transformer_input(
        source_entity: RulesEntity, transformer: VerifiedRulesTransformer
    ) -> InformationRules | DMSRules:
        # Case 1: We only have information rules
        if source_entity.dms is None:
            if transformer.is_valid_input(source_entity.information):
                return source_entity.information
            raise InvalidActivityInput(expected=(DMSRules,), have=(InformationRules,))
        # Case 2: We have both information and dms rules and the transformer is compatible with dms rules
        elif isinstance(source_entity.dms, DMSRules) and transformer.is_valid_input(source_entity.dms):
            return source_entity.dms
        # Case 3: We have both information and dms rules and the transformer is compatible with information rules
        raise InvalidActivityInput(expected=(InformationRules,), have=(DMSRules,))

    def _get_source_id(self, result: tuple[InformationRules, DMSRules | None]) -> rdflib.URIRef | None:
        """Return the source of the result.

        !!! note
            This method prioritizes the source_id of the DMS rules
        """
        info, dms = result
        return dms.metadata.source_id if dms else info.metadata.source_id

    def _create_id(self, info: InformationRules, dms: DMSRules | None) -> rdflib.URIRef:
        if dms is None:
            identifier = info.metadata.identifier
        else:
            identifier = dms.metadata.identifier

        # Here we check if the identifier is already in the iteration dictionary
        # to track specific changes to the same entity, if it is we increment the iteration
        if identifier not in self._iteration_by_id:
            self._iteration_by_id[identifier] = 1
            return identifier

        # If the identifier is already in the iteration dictionary we increment the iteration
        # and update identifier to include the iteration number
        self._iteration_by_id[identifier] += 1
        return identifier + f"/Iteration_{self._iteration_by_id[identifier]}"

    @property
    def last_verified_dms_rules(self) -> DMSRules:
        if not self.provenance:
            raise EmptyStore()
        if self.provenance[-1].target_entity.dms is None:
            raise NeatValueError("No verified DMS rules found in the provenance.")
        return self.provenance[-1].target_entity.dms

    @property
    def last_verified_information_rules(self) -> InformationRules:
        if not self.provenance:
            raise EmptyStore()
        return self.provenance[-1].target_entity.information

    @property
    def last_issues(self) -> IssueList | None:
        return self._last_issues

    @property
    def last_outcome(self) -> UploadResultList:
        if self._last_outcome is not None:
            return self._last_outcome
        raise NeatValueError("No outcome found in the provenance.")

    @property
    def empty(self) -> bool:
        """Check if the store is empty."""
        return not self.provenance



from ._constants import Undefined, Unknown
from ._loaders import load_connection, load_dms_value_type, load_value_type
from ._multi_value import MultiValueTypeInfo
from ._single_value import (
    AssetEntity,
    AssetFields,
    ClassEntity,
    ContainerEntity,
    DataModelEntity,
    DMSEntity,
    DMSNodeEntity,
    DMSUnknownEntity,
    DMSVersionedEntity,
    EdgeEntity,
    Entity,
    ReferenceEntity,
    RelationshipEntity,
    ReverseConnectionEntity,
    T_Entity,
    UnitEntity,
    UnknownEntity,
    ViewEntity,
)
from ._types import CdfResourceEntityList, ClassEntityList, ContainerEntityList, URLEntity, ViewEntityList
from ._wrapped import DMSFilter, HasDataFilter, NodeTypeFilter, RawFilter, WrappedEntity

__all__ = [
    "AssetEntity",
    "AssetFields",
    "CdfResourceEntityList",
    "ClassEntity",
    "ClassEntityList",
    "ContainerEntity",
    "ContainerEntityList",
    "DMSEntity",
    "DMSFilter",
    "DMSNodeEntity",
    "DMSUnknownEntity",
    "DMSVersionedEntity",
    "DataModelEntity",
    "EdgeEntity",
    "Entity",
    "HasDataFilter",
    "MultiValueTypeInfo",
    "NodeTypeFilter",
    "RawFilter",
    "ReferenceEntity",
    "RelationshipEntity",
    "ReverseConnectionEntity",
    "T_Entity",
    "URLEntity",
    "Undefined",
    "UnitEntity",
    "Unknown",
    "UnknownEntity",
    "ViewEntity",
    "ViewEntityList",
    "WrappedEntity",
    "load_connection",
    "load_dms_value_type",
    "load_value_type",
]



from typing import cast
from warnings import WarningMessage

from pydantic_core import ErrorDetails

from cognite.neat._issues._base import NeatError, NeatWarning
from cognite.neat._utils.spreadsheet import SpreadsheetRead

from .errors import NeatValueError, SpreadsheetError
from .warnings import NeatValueWarning


def from_pydantic_errors(
    errors: list[ErrorDetails], read_info_by_sheet: dict[str, SpreadsheetRead] | None = None
) -> list[NeatError]:
    read_info_by_sheet = read_info_by_sheet or {}
    return [
        _from_pydantic_error(error, read_info_by_sheet)
        for error in errors
        # Skip the error for SheetList, as it is not relevant for the user. This is an
        # internal class used to have helper methods for a lists as .to_pandas()
        if not (error["type"] == "is_instance_of" and error["loc"][1] == "is-instance[SheetList]")
    ]


def from_warning(warning: WarningMessage) -> NeatWarning:
    if isinstance(warning.message, NeatWarning):
        return warning.message
    message = f"{warning.category.__name__}: {warning.message!s}"
    if warning.source:
        message += f" Source: {warning.source}"
    return NeatValueWarning(message)


def _from_pydantic_error(error: ErrorDetails, read_info_by_sheet: dict[str, SpreadsheetRead]) -> NeatError:
    neat_error = _create_neat_value_error(error)
    location = error["loc"]

    # only errors caused in model_validate will have location information
    if location:
        return SpreadsheetError.create(location, neat_error, read_info_by_sheet.get(cast(str, location[0])))

    # errors that occur while for example parsing spreadsheet in input rules
    # will not have location information so we return neat_error as is
    # this is workaround until more elegant solution is found
    return neat_error


def _create_neat_value_error(error: ErrorDetails) -> NeatValueError:
    if (ctx := error.get("ctx")) and (neat_error := ctx.get("error")) and isinstance(neat_error, NeatError):
        # Is already a NeatError
        return neat_error
    return _pydantic_to_neat_error(error)


def _pydantic_to_neat_error(error: ErrorDetails) -> NeatValueError:
    error_type = error["type"]
    input_value = error["input"]
    match error_type:
        # See https://docs.pydantic.dev/latest/errors/validation_errors/ for all possible error types:
        case error_type if error_type.endswith("_type") | error_type.endswith("_parsing"):
            if input_value is None:
                return NeatValueError("value is missing.")
            expected_type = error_type.removesuffix("_type").removesuffix("_parsing")
            return NeatValueError(f"Expected a {expected_type} type, got {input_value!r}")
        case _:
            # The above cases overwrite the human-readable message from pydantic.
            # Motivation for overwriting is that pydantic is developer-oriented and while neat is SME-oriented.
            return NeatValueError(f"{error['msg']} got '{input_value}'")



from abc import ABC, abstractmethod
from contextlib import suppress
from datetime import datetime
from typing import TYPE_CHECKING, Any, Generic

from rdflib import URIRef

from cognite.neat._constants import DEFAULT_NAMESPACE
from cognite.neat._rules._shared import ReadRules, T_InputRules
from cognite.neat._utils.auxiliary import class_html_doc

if TYPE_CHECKING:
    from cognite.neat._store._provenance import Agent as ProvenanceAgent


class BaseImporter(ABC, Generic[T_InputRules]):
    """
    BaseImporter class which all importers inherit from.
    """

    @abstractmethod
    def to_rules(self) -> ReadRules[T_InputRules]:
        """Creates `Rules` object from the data for target role."""
        raise NotImplementedError()

    def _default_metadata(self) -> dict[str, Any]:
        creator = "UNKNOWN"
        with suppress(KeyError, ImportError):
            import getpass

            creator = getpass.getuser()

        return {
            "schema": "partial",
            "space": "neat",
            "external_id": "NeatImportedDataModel",
            "version": "0.1.0",
            "name": "Neat Imported Data Model",
            "created": datetime.now().replace(microsecond=0).isoformat(),
            "updated": datetime.now().replace(microsecond=0).isoformat(),
            "creator": creator,
            "description": f"Imported using {type(self).__name__}",
        }

    @classmethod
    def _repr_html_(cls) -> str:
        return class_html_doc(cls)

    @property
    def agent(self) -> "ProvenanceAgent":
        """Provenance agent for the importer."""
        from cognite.neat._store._provenance import Agent as ProvenanceAgent

        return ProvenanceAgent(id_=DEFAULT_NAMESPACE[f"agent/{type(self).__name__}"])

    @property
    def description(self) -> str:
        return "MISSING DESCRIPTION"

    @property
    def source_uri(self) -> URIRef:
        return DEFAULT_NAMESPACE["UNKNOWN"]



from ._base import BaseImporter
from ._dms2rules import DMSImporter
from ._dtdl2rules import DTDLImporter
from ._rdf import IMFImporter, InferenceImporter, OWLImporter, SubclassInferenceImporter
from ._spreadsheet2rules import ExcelImporter, GoogleSheetImporter
from ._yaml2rules import YAMLImporter

__all__ = [
    "BaseImporter",
    "DMSImporter",
    "DTDLImporter",
    "ExcelImporter",
    "GoogleSheetImporter",
    "IMFImporter",
    "InferenceImporter",
    "OWLImporter",
    "SubclassInferenceImporter",
    "YAMLImporter",
]

RulesImporters = (
    OWLImporter
    | IMFImporter
    | DMSImporter
    | ExcelImporter
    | GoogleSheetImporter
    | DTDLImporter
    | YAMLImporter
    | InferenceImporter
    | SubclassInferenceImporter
)


def _repr_html_() -> str:
    import pandas as pd

    table = pd.DataFrame(  # type: ignore[operator]
        [
            {
                "Importer": name,
                "Description": (
                    globals()[name].__doc__.strip().split("\n")[0] if globals()[name].__doc__ else "Missing"
                ),
            }
            for name in __all__
            if name != "BaseImporter"
        ]
    )._repr_html_()

    return (
        "<strong>Importer</strong> An importer reads data/schema/data model from a source"
        f" and converts it into Neat's representation of a data model called <em>Rules</em>.<br />{table}"
    )



import hashlib
import importlib
import inspect
import logging
import time
from collections.abc import Callable
from datetime import datetime
from functools import wraps
from types import ModuleType

from cognite.client.exceptions import CogniteDuplicatedError, CogniteReadTimeout

from cognite.neat._issues.errors import NeatImportError


def local_import(module: str, extra: str) -> ModuleType:
    try:
        return importlib.import_module(module)
    except ImportError as e:
        raise NeatImportError(module.split(".")[0], extra) from e


def get_classmethods(cls: type) -> list[Callable]:
    return [
        func for _, func in inspect.getmembers(cls, lambda x: inspect.ismethod(x) and not x.__name__.startswith("_"))
    ]


def class_html_doc(cls: type, include_factory_methods: bool = True) -> str:
    if cls.__doc__:
        docstring = cls.__doc__.split("Args:")[0].strip().replace("\n", "<br />")
    else:
        docstring = "Missing Description"
    if include_factory_methods:
        factory_methods = get_classmethods(cls)
        if factory_methods:
            factory_methods_str = "".join(f"<li><em>.{m.__name__}</em></li>" for m in factory_methods)
            docstring += (
                f"<br /><strong>Available factory methods:</strong><br />"
                f'<ul style="list-style-type:circle;">{factory_methods_str}</ul>'
            )
    return f"<h3>{cls.__name__}</h3><p>{docstring}</p>"


def retry_decorator(max_retries=2, retry_delay=3, component_name=""):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            previous_exception = None
            for attempt in range(max_retries + 1):
                try:
                    logging.debug(f"Attempt {attempt + 1} of {max_retries + 1} for {component_name}")
                    return func(*args, **kwargs)
                except CogniteReadTimeout as e:
                    previous_exception = e
                    if attempt < max_retries:
                        logging.error(
                            f"""CogniteReadTimeout retry attempt {attempt + 1} failed for {component_name} .
                            Retrying in {retry_delay} second(s). Error:"""
                        )
                        logging.error(e)
                        time.sleep(retry_delay)
                    else:
                        raise e
                except CogniteDuplicatedError as e:
                    if isinstance(previous_exception, CogniteReadTimeout):
                        # if previous exception was CogniteReadTimeout,
                        # we can't be sure if the items were created or not
                        if len(e.successful) == 0 and len(e.failed) == 0 and len(e.duplicated) >= 0:
                            logging.warning(
                                f"Duplicate error for {component_name} . All items already exist in CDF. "
                                "Suppressing error."
                            )
                            return
                        else:
                            # can happend because of eventual consistency. Retry with delay to allow for CDF to catch up
                            if attempt < max_retries:
                                logging.error(
                                    f"""CogniteDuplicatedError retry attempt {attempt + 1} failed for {component_name} .
                                      Retrying in {retry_delay} second(s). Error:"""
                                )
                                logging.error(e)
                                # incerasing delay to allow for CDF to catch up
                                time.sleep(retry_delay)
                            else:
                                raise e
                    else:
                        # no point in retrying duplicate error if previous exception was not a timeout
                        raise e

                except Exception as e:
                    previous_exception = e
                    if attempt < max_retries:
                        logging.error(
                            f"Retry attempt {attempt + 1} failed for {component_name}. "
                            f"Retrying in {retry_delay} second(s)."
                        )
                        logging.error(e)
                        time.sleep(retry_delay)
                    else:
                        raise e

        return wrapper

    return decorator


def create_sha256_hash(string: str) -> str:
    # Create a SHA-256 hash object
    sha256_hash = hashlib.sha256()

    # Convert the string to bytes and update the hash object
    sha256_hash.update(string.encode("utf-8"))

    # Get the hexadecimal representation of the hash
    hash_value = sha256_hash.hexdigest()

    return hash_value


def string_to_ideal_type(input_string: str) -> int | bool | float | datetime | str:
    try:
        # Try converting to int
        return int(input_string)
    except ValueError:
        try:
            # Try converting to float
            return float(input_string)  # type: ignore
        except ValueError:
            if input_string.lower() == "true":
                # Return True if input is 'true'
                return True
            elif input_string.lower() == "false":
                # Return False if input is 'false'
                return False
            else:
                try:
                    # Try converting to datetime
                    return datetime.fromisoformat(input_string)  # type: ignore
                except ValueError:
                    # Return the input string if no conversion is possible
                    return input_string


def get_parameters_by_method(obj: object, prefix: str = "") -> dict[str, dict[str, type]]:
    annotation_by_method = {}
    namespace_dict = {**dict(obj.__class__.__dict__.items()), **vars(obj)}
    for name, value in namespace_dict.items():
        if name != "__call__" and (name.startswith("_") or isinstance(value, property)):
            continue
        if callable(value) and type(value).__name__ == "function":
            annotation_by_method[f"{prefix}{name}"] = get_parameters(value)
        elif isinstance(value, object) and type(value).__module__ != "builtins":
            sub_prefix = f"{prefix}{name}." if prefix else f"{name}."
            annotation_by_method.update(get_parameters_by_method(value, sub_prefix))
    return annotation_by_method


def get_parameters(obj: Callable) -> dict[str, type]:
    annotations = inspect.get_annotations(obj)
    return {name: annotations[name] for name in annotations if name != "return"}



import typing
import urllib.parse
import warnings
from collections import defaultdict
from collections.abc import Iterable, Sequence
from typing import ClassVar, NamedTuple, cast

from cognite.client import CogniteClient
from cognite.client.exceptions import CogniteAPIError
from rdflib import Namespace, URIRef

from cognite.neat._constants import CLASSIC_CDF_NAMESPACE, DEFAULT_NAMESPACE, get_default_prefixes_and_namespaces
from cognite.neat._graph.extractors._base import KnowledgeGraphExtractor
from cognite.neat._issues.errors import NeatValueError, ResourceNotFoundError
from cognite.neat._issues.warnings import CDFAuthWarning, NeatValueWarning
from cognite.neat._rules._shared import ReadRules
from cognite.neat._rules.catalog import classic_model
from cognite.neat._rules.models import InformationInputRules, InformationRules
from cognite.neat._shared import Triple
from cognite.neat._utils.collection_ import chunker, iterate_progress_bar
from cognite.neat._utils.rdf_ import remove_namespace_from_uri
from cognite.neat._utils.text import to_snake_case

from ._assets import AssetsExtractor
from ._base import ClassicCDFBaseExtractor, InstanceIdPrefix
from ._data_sets import DataSetExtractor
from ._events import EventsExtractor
from ._files import FilesExtractor
from ._labels import LabelsExtractor
from ._relationships import RelationshipsExtractor
from ._sequences import SequencesExtractor
from ._timeseries import TimeSeriesExtractor


class _ClassicCoreType(NamedTuple):
    extractor_cls: (
        type[AssetsExtractor]
        | type[TimeSeriesExtractor]
        | type[SequencesExtractor]
        | type[EventsExtractor]
        | type[FilesExtractor]
    )
    resource_type: InstanceIdPrefix
    api_name: str


class ClassicGraphExtractor(KnowledgeGraphExtractor):
    """This extractor extracts all classic CDF Resources.

    The Classic Graph consists of the following core resource type.

    Classic Node CDF Resources:
     - Assets
     - TimeSeries
     - Sequences
     - Events
     - Files

    All the classic node CDF resources can have one or more connections to one or more assets. This
    will match a direct relationship in the data modeling of CDF.

    In addition, you have relationships between the classic node CDF resources. This matches an edge
    in the data modeling of CDF.

    Finally, you have labels and data sets that to organize the graph. In which data sets have a similar,
    but different, role as a space in data modeling. While labels can be compared to node types in data modeling,
    used to quickly filter and find nodes/edges.

    This extractor will extract the classic CDF graph into Neat starting from either a data set or a root asset.

    It works as follows:

    1. Extract all core nodes (assets, time series, sequences, events, files) filtered by the given data set or
       root asset.
    2. Extract all relationships starting from any of the extracted core nodes.
    3. Extract all core nodes that are targets of the relationships that are not already extracted.
    4. Extract all labels that are connected to the extracted core nodes/relationships.
    5. Extract all data sets that are connected to the extracted core nodes/relationships.

    Args:
        client (CogniteClient): The Cognite client to use.
        data_set_external_id (str, optional): The data set external id to extract from. Defaults to None.
        root_asset_external_id (str, optional): The root asset external id to extract from. Defaults to None.
        namespace (Namespace, optional): The namespace to use. Defaults to DEFAULT_NAMESPACE.
    """

    # These are the core resource types in the classic CDF.
    _classic_node_types: ClassVar[tuple[_ClassicCoreType, ...]] = (
        _ClassicCoreType(AssetsExtractor, InstanceIdPrefix.asset, "assets"),
        _ClassicCoreType(TimeSeriesExtractor, InstanceIdPrefix.time_series, "time_series"),
        _ClassicCoreType(SequencesExtractor, InstanceIdPrefix.sequence, "sequences"),
        _ClassicCoreType(EventsExtractor, InstanceIdPrefix.event, "events"),
        _ClassicCoreType(FilesExtractor, InstanceIdPrefix.file, "files"),
    )

    def __init__(
        self,
        client: CogniteClient,
        data_set_external_id: str | None = None,
        root_asset_external_id: str | None = None,
        namespace: Namespace | None = None,
        limit_per_type: int | None = None,
        prefix: str | None = None,
        identifier: typing.Literal["id", "externalId"] = "id",
        unpack_metadata: bool = False,
        skip_sequence_rows: bool = False,
    ):
        self._client = client
        if sum([bool(data_set_external_id), bool(root_asset_external_id)]) != 1:
            raise ValueError("Exactly one of data_set_external_id or root_asset_external_id must be set.")
        self._root_asset_external_id = root_asset_external_id
        self._data_set_external_id = data_set_external_id
        self._namespace = namespace or CLASSIC_CDF_NAMESPACE
        self._extractor_args = dict(
            namespace=self._namespace,
            unpack_metadata=unpack_metadata,
            as_write=True,
            camel_case=True,
            limit=limit_per_type,
            prefix=prefix,
            identifier=identifier,
        )
        self._identifier = identifier
        self._prefix = prefix
        self._limit_per_type = limit_per_type
        self._skip_sequence_rows = skip_sequence_rows

        self._uris_by_external_id_by_type: dict[InstanceIdPrefix, dict[str, URIRef]] = defaultdict(dict)
        self._source_external_ids_by_type: dict[InstanceIdPrefix, set[str]] = defaultdict(set)
        self._target_external_ids_by_type: dict[InstanceIdPrefix, set[str]] = defaultdict(set)
        self._relationship_subject_predicate_type_external_id: list[tuple[URIRef, URIRef, str, str]] = []
        self._labels: set[str] = set()
        self._data_set_ids: set[int] = set()
        self._data_set_external_ids: set[str] = set()
        self._extracted_labels = False
        self._extracted_data_sets = False
        self._asset_external_ids_by_id: dict[int, str] = {}
        self._dataset_external_ids_by_id: dict[int, str] = {}
        self.neat_prefix_by_predicate_uri: dict[URIRef, str] = {
            self._namespace["dataSetId"]: InstanceIdPrefix.data_set,
            self._namespace["assetId"]: InstanceIdPrefix.asset,
            self._namespace["assetIds"]: InstanceIdPrefix.asset,
            self._namespace["parentId"]: InstanceIdPrefix.asset,
            self._namespace["rootId"]: InstanceIdPrefix.asset,
            self._namespace["labels"]: InstanceIdPrefix.label,
        }
        self.neat_prefix_by_type_uri: dict[URIRef, str] = {}
        for extractor_cls, type_prefix in [
            (AssetsExtractor, InstanceIdPrefix.asset),
            (TimeSeriesExtractor, InstanceIdPrefix.time_series),
            (SequencesExtractor, InstanceIdPrefix.sequence),
            (EventsExtractor, InstanceIdPrefix.event),
            (FilesExtractor, InstanceIdPrefix.file),
            (RelationshipsExtractor, InstanceIdPrefix.relationship),
            (LabelsExtractor, InstanceIdPrefix.label),
            (DataSetExtractor, InstanceIdPrefix.data_set),
        ]:
            rdf_type = extractor_cls._default_rdf_type  # type: ignore[attr-defined]
            if prefix:
                rdf_type = f"{prefix}{rdf_type}"
            self.neat_prefix_by_type_uri[self._namespace[rdf_type]] = type_prefix

    def _get_activity_names(self) -> list[str]:
        activities = [data_access_object.extractor_cls.__name__ for data_access_object in self._classic_node_types] + [
            RelationshipsExtractor.__name__,
        ]
        if self._extracted_labels:
            activities.append(LabelsExtractor.__name__)
        if self._extracted_data_sets:
            activities.append(DataSetExtractor.__name__)
        return activities

    def extract(self) -> Iterable[Triple]:
        """Extracts all classic CDF Resources."""
        self._validate_exists()

        yield from self._extract_core_start_nodes()

        yield from self._extract_start_node_relationships()

        yield from self._extract_core_end_nodes()

        if self._identifier == "id":
            yield from self._extract_relationship_target_triples()

        try:
            yield from self._extract_labels()
        except CogniteAPIError as e:
            warnings.warn(CDFAuthWarning("extract labels", str(e)), stacklevel=2)
        else:
            self._extracted_labels = True

        try:
            yield from self._extract_data_sets()
        except CogniteAPIError as e:
            warnings.warn(CDFAuthWarning("extract data sets", str(e)), stacklevel=2)
        else:
            self._extracted_data_sets = True

    def get_information_rules(self) -> InformationRules:
        # To avoid circular imports
        from cognite.neat._rules.importers import ExcelImporter

        unverified = cast(ReadRules[InformationInputRules], ExcelImporter(classic_model).to_rules())
        if unverified.rules is None:
            raise NeatValueError(f"Could not read the classic model rules from {classic_model}.")

        verified = unverified.rules.as_verified_rules()
        prefixes = get_default_prefixes_and_namespaces()
        instance_prefix: str | None = next((k for k, v in prefixes.items() if v == self._namespace), None)
        if instance_prefix is None:
            # We need to add a new prefix
            instance_prefix = f"prefix_{len(prefixes) + 1}"
            prefixes[instance_prefix] = self._namespace
        verified.prefixes = prefixes

        is_snake_case = self._extractor_args["camel_case"] is False
        for prop in verified.properties:
            prop_id = prop.property_
            if is_snake_case:
                prop_id = to_snake_case(prop_id)
            prop.instance_source = [self._namespace[prop_id]]
        for cls_ in verified.classes:
            cls_id = cls_.class_.suffix
            if is_snake_case:
                cls_id = to_snake_case(cls_id)
            cls_.instance_source = self._namespace[cls_id]

        return verified

    @property
    def description(self) -> str:
        if self._data_set_external_id:
            source = f"data set {self._data_set_external_id}."
        elif self._root_asset_external_id:
            source = f"root asset {self._root_asset_external_id}."
        else:
            source = "unknown source."
        return f"Extracting clasic CDF Graph (Assets, TimeSeries, Sequences, Events, Files) from {source}."

    @property
    def source_uri(self) -> URIRef:
        if self._data_set_external_id:
            resource = "dataset"
            external_id = self._data_set_external_id
        elif self._root_asset_external_id:
            resource = "asset"
            external_id = self._root_asset_external_id
        else:
            resource = "unknown"
            external_id = "unknown"
        return DEFAULT_NAMESPACE[f"{self._client.config.project}/{resource}/{urllib.parse.quote(external_id)}"]

    def _validate_exists(self) -> None:
        if self._data_set_external_id:
            if self._client.data_sets.retrieve(external_id=self._data_set_external_id) is None:
                raise ResourceNotFoundError(self._data_set_external_id, "data set")
        elif self._root_asset_external_id:
            if self._client.assets.retrieve(external_id=self._root_asset_external_id) is None:
                raise ResourceNotFoundError(self._root_asset_external_id, "root asset")
        else:
            raise ValueError("Exactly one of data_set_external_id or root_asset_external_id must be set.")

    def _extract_core_start_nodes(self):
        for core_node in self._classic_node_types:
            kwargs = self._extractor_args.copy()
            if core_node.extractor_cls == SequencesExtractor and self._skip_sequence_rows:
                kwargs["skip_rows"] = True

            if self._data_set_external_id:
                extractor = core_node.extractor_cls.from_dataset(self._client, self._data_set_external_id, **kwargs)
            elif self._root_asset_external_id:
                extractor = core_node.extractor_cls.from_hierarchy(self._client, self._root_asset_external_id, **kwargs)
            else:
                raise ValueError("Exactly one of data_set_external_id or root_asset_external_id must be set.")

            if self._identifier == "externalId":
                if isinstance(extractor, AssetsExtractor):
                    self._asset_external_ids_by_id = extractor.asset_external_ids_by_id
                else:
                    extractor.asset_external_ids_by_id = self._asset_external_ids_by_id
                extractor.lookup_dataset_external_id = self._lookup_dataset
            elif self._identifier == "id":
                extractor._log_urirefs = True

            yield from self._extract_with_logging_label_dataset(extractor, core_node.resource_type)

            if self._identifier == "id":
                self._uris_by_external_id_by_type[core_node.resource_type].update(extractor._uriref_by_external_id)

    def _extract_start_node_relationships(self):
        for start_resource_type, source_external_ids in self._source_external_ids_by_type.items():
            start_type = start_resource_type.removesuffix("_")
            for chunk in self._chunk(list(source_external_ids), description=f"Extracting {start_type} relationships"):
                relationship_iterator = self._client.relationships(
                    source_external_ids=list(chunk), source_types=[start_type]
                )
                extractor = RelationshipsExtractor(relationship_iterator, **self._extractor_args)
                # This is a private attribute, but we need to set it to log the target nodes.
                extractor._log_target_nodes = True
                if self._identifier == "id":
                    extractor._uri_by_external_id_by_type = self._uris_by_external_id_by_type
                elif self._identifier == "externalId":
                    extractor.lookup_dataset_external_id = self._lookup_dataset

                yield from extractor.extract()

                # After the extraction is done, we need to update all the new target nodes so
                # we can extract them in the next step.
                for end_type, target_external_ids in extractor._target_external_ids_by_type.items():
                    for external_id in target_external_ids:
                        # We only want to extract the target nodes that are not already extracted.
                        # Even though _source_external_ids_by_type is a defaultdict, we have to check if the key exists.
                        # This is because we might not have extracted any nodes of that type yet, and looking up
                        # a key that does not exist will create it. We are iterating of this dictionary, and
                        # we do not want to create new keys while iterating.
                        if (
                            end_type not in self._source_external_ids_by_type
                            or external_id not in self._source_external_ids_by_type[end_type]
                        ):
                            self._target_external_ids_by_type[end_type].add(external_id)

                if self._identifier == "id":
                    # We need to store all future target triples which we will lookup after fetching
                    # the target nodes.
                    self._relationship_subject_predicate_type_external_id.extend(extractor._target_triples)

    def _extract_core_end_nodes(self):
        for core_node in self._classic_node_types:
            target_external_ids = self._target_external_ids_by_type[core_node.resource_type]
            api = getattr(self._client, core_node.api_name)
            for chunk in self._chunk(
                list(target_external_ids),
                description=f"Extracting end nodes {core_node.resource_type.removesuffix('_')}",
            ):
                resource_iterator = api.retrieve_multiple(external_ids=list(chunk), ignore_unknown_ids=True)
                extractor = core_node.extractor_cls(resource_iterator, **self._extractor_args)

                extractor.asset_external_ids_by_id = self._asset_external_ids_by_id
                extractor.lookup_dataset_external_id = self._lookup_dataset
                if self._identifier == "id":
                    extractor._log_urirefs = True

                yield from self._extract_with_logging_label_dataset(extractor)

                if self._identifier == "id":
                    self._uris_by_external_id_by_type[core_node.resource_type].update(extractor._uriref_by_external_id)

    def _extract_relationship_target_triples(self):
        for id_, predicate, type_, external_id in self._relationship_subject_predicate_type_external_id:
            try:
                object_uri = self._uris_by_external_id_by_type[InstanceIdPrefix.from_str(type_)][external_id]
            except KeyError:
                warnings.warn(NeatValueWarning(f"Missing externalId {external_id} for {type_}"), stacklevel=2)
            else:
                yield id_, predicate, object_uri

    def _extract_labels(self):
        for chunk in self._chunk(list(self._labels), description="Extracting labels"):
            label_iterator = self._client.labels.retrieve(external_id=list(chunk), ignore_unknown_ids=True)
            extractor = LabelsExtractor(label_iterator, **self._extractor_args)
            extractor.lookup_dataset_external_id = self._lookup_dataset
            yield from extractor.extract()

    def _extract_data_sets(self):
        for chunk in self._chunk(list(self._data_set_ids), description="Extracting data sets"):
            data_set_iterator = self._client.data_sets.retrieve_multiple(ids=list(chunk), ignore_unknown_ids=True)
            yield from DataSetExtractor(data_set_iterator, **self._extractor_args).extract()
        for chunk in self._chunk(list(self._data_set_external_ids), description="Extracting data sets"):
            data_set_iterator = self._client.data_sets.retrieve_multiple(
                external_ids=list(chunk), ignore_unknown_ids=True
            )
            yield from DataSetExtractor(data_set_iterator, **self._extractor_args).extract()

    def _extract_with_logging_label_dataset(
        self, extractor: ClassicCDFBaseExtractor, resource_type: InstanceIdPrefix | None = None
    ) -> Iterable[Triple]:
        for triple in extractor.extract():
            if triple[1] == self._namespace.externalId and resource_type is not None:
                self._source_external_ids_by_type[resource_type].add(remove_namespace_from_uri(triple[2]))
            elif triple[1] == self._namespace.labels:
                self._labels.add(remove_namespace_from_uri(triple[2]).removeprefix(InstanceIdPrefix.label))
            elif triple[1] == self._namespace.dataSetId:
                identifier = remove_namespace_from_uri(triple[2]).removeprefix(InstanceIdPrefix.data_set)
                try:
                    self._data_set_ids.add(int(identifier))
                except ValueError:
                    self._data_set_external_ids.add(identifier)
            yield triple

    @staticmethod
    def _chunk(items: Sequence, description: str) -> Iterable:
        to_iterate: Iterable = chunker(items, chunk_size=1000)
        if items:
            return iterate_progress_bar(to_iterate, (len(items) // 1_000) + 1, description)
        else:
            return to_iterate

    def _lookup_dataset(self, dataset_id: int) -> str:
        if dataset_id not in self._dataset_external_ids_by_id:
            try:
                if (dataset := self._client.data_sets.retrieve(id=dataset_id)) and dataset.external_id:
                    self._dataset_external_ids_by_id[dataset_id] = dataset.external_id
                else:
                    raise KeyError(f"Could not find dataset with id {dataset_id}.")
            except CogniteAPIError as e:
                warnings.warn(CDFAuthWarning("lookup dataset", str(e)), stacklevel=2)
                return f"{InstanceIdPrefix.data_set}{dataset_id}"
        return self._dataset_external_ids_by_id[dataset_id]



from abc import ABC
from dataclasses import dataclass
from typing import ClassVar

from cognite.client.data_classes.data_modeling import DataModelId, ViewId

from cognite.neat._constants import DMS_VIEW_CONTAINER_SIZE_LIMIT
from cognite.neat._issues import NeatWarning

_BASE_URL = "https://cognite-neat.readthedocs-hosted.com/en/latest/data-modeling-principles.html"


@dataclass(unsafe_hash=True)
class BreakingModelingPrincipleWarning(NeatWarning, ABC):
    """{warning_class}: {specific} violates the {principle} principle.
    See {url} for more information."""

    url: ClassVar[str]
    specific: str

    def as_message(self, include_type: bool = True) -> str:
        principle = type(self).__name__.removesuffix("Warning")
        url = f"{_BASE_URL}#{self.url}"
        return (self.__doc__ or "").format(
            warning_class=BreakingModelingPrincipleWarning.__name__,
            specific=self.specific,
            principle=principle,
            url=url,
        )


@dataclass(unsafe_hash=True)
class PrincipleOneModelOneSpaceWarning(BreakingModelingPrincipleWarning):
    """{warning_class}: {specific} violates the {principle} principle.
    See {url} for more information."""

    url = "all-data-models-are-kept-in-its-own-space"


@dataclass(unsafe_hash=True)
class PrincipleMatchingSpaceAndVersionWarning(BreakingModelingPrincipleWarning):
    """{warning_class}: {specific} violates the {principle} principle.
    See {url} for more information."""

    url = "all-views-of-a-data-models-have-the-same-version-and-space-as-the-data-model"


@dataclass(unsafe_hash=True)
class PrincipleSolutionBuildsOnEnterpriseWarning(BreakingModelingPrincipleWarning):
    """{warning_class}: {specific} violates the {principle} principle.
    See {url} for more information."""

    url = "solution-data-models-should-always-be-referencing-the-enterprise-data-model"


@dataclass(unsafe_hash=True)
class UserModelingWarning(NeatWarning, ABC):
    """This is a generic warning for user modeling issues.
    These warnings will not cause the resulting model to be invalid, but
    will likely lead to suboptimal performance, unnecessary complexity, or other issues."""

    ...


@dataclass(unsafe_hash=True)
class CDFNotSupportedWarning(NeatWarning, ABC):
    """This is a base class for warnings for modeling issues that will
    likely lead to the CDF API rejecting the model."""

    ...


@dataclass(unsafe_hash=True)
class NotSupportedViewContainerLimitWarning(CDFNotSupportedWarning):
    """The view {view_id} maps, {count} containers, which is more than the limit {limit}."""

    fix = "Reduce the number of containers the view maps to."

    view_id: ViewId
    count: int
    limit: int = DMS_VIEW_CONTAINER_SIZE_LIMIT


@dataclass(unsafe_hash=True)
class NotSupportedHasDataFilterLimitWarning(CDFNotSupportedWarning):
    """The view {view_id} uses a hasData filter applied to {count} containers, which is more than the limit {limit}."""

    fix = "Do not map to more than {limit} containers."

    view_id: ViewId
    count: int
    limit: int = DMS_VIEW_CONTAINER_SIZE_LIMIT


@dataclass(unsafe_hash=True)
class UndefinedClassWarning(UserModelingWarning):
    """Class {class_id} has no explicit properties defined neither implements other class"""

    fix = "Define properties for class or inherit properties by implementing another class."

    class_id: str


@dataclass(unsafe_hash=True)
class UndefinedViewWarning(UserModelingWarning):
    """Undefined view {value_type} has been referred as value type for property <{view_property}> of view {view_id}."""

    fix = "Define views which are used as value types."

    view_id: str
    value_type: str
    view_property: str


@dataclass(unsafe_hash=True)
class EnterpriseModelNotBuildOnTopOfCDMWarning(UserModelingWarning):
    """Enterprise data model being build on top {reference_model_id}. This is not recommended."""

    fix = "Always build Enterprise Data Model on top of Cognite Data Model such as Core Data Model."

    reference_model_id: DataModelId


@dataclass(unsafe_hash=True)
class SolutionModelBuildOnTopOfCDMWarning(UserModelingWarning):
    """Solution data model being build on top Cognite Data Model {reference_model_id}. This is not recommended."""

    fix = "Always build solution data model on top of enterprise data model."

    reference_model_id: DataModelId



"""This module performs importing of various formats to one of serializations for which
there are loaders to TransformationRules pydantic class."""

from cognite.neat._rules.importers._rdf._base import BaseRDFImporter
from cognite.neat._rules.importers._rdf._shared import parse_classes, parse_properties

CLASSES_QUERY = """SELECT ?class_ ?name ?description ?implements
        WHERE {{

        ?class_ a owl:Class .
        OPTIONAL {{?class_ rdfs:subClassOf ?implements }}.
        OPTIONAL {{?class_ rdfs:label|skos:prefLabel ?name }}.
        OPTIONAL {{?class_ rdfs:comment|skos:definition ?description}} .


        FILTER (!isBlank(?class_))
        FILTER (!bound(?implements) || !isBlank(?implements))

        FILTER (!bound(?name) || LANG(?name) = "" || LANGMATCHES(LANG(?name), "{language}"))
        FILTER (!bound(?description) || LANG(?description) = "" || LANGMATCHES(LANG(?description), "{language}"))

    }}
    """

PROPERTIES_QUERY = """

    SELECT ?class_ ?property_ ?name ?description ?value_type ?minCount ?maxCount ?default
    WHERE {{
        ?property_ a ?property_Type.
        FILTER (?property_Type IN (owl:ObjectProperty, owl:DatatypeProperty ) )
        OPTIONAL {{?property_ rdfs:domain ?class_ }}.
        OPTIONAL {{?property_ rdfs:range ?value_type }}.
        OPTIONAL {{?property_ rdfs:label|skos:prefLabel ?name }}.
        OPTIONAL {{?property_ rdfs:comment|skos:definition ?description}}.
        OPTIONAL {{?property_ owl:maxCardinality ?maxCount}}.
        OPTIONAL {{?property_ owl:minCardinality ?minCount}}.

        # FILTERS
        FILTER (!isBlank(?property_))
        FILTER (!bound(?name) || LANG(?name) = "" || LANGMATCHES(LANG(?name), "{language}"))
        FILTER (!bound(?description) || LANG(?description) = "" || LANGMATCHES(LANG(?description), "{language}"))
    }}
    """


class OWLImporter(BaseRDFImporter):
    """Convert OWL ontology to tables/ transformation rules / Excel file.

        Args:
            filepath: Path to OWL ontology

    !!! Note
        OWL Ontologies are information models which completeness varies. As such, constructing functional
        data model directly will often be impossible, therefore the produced Rules object will be ill formed.
        To avoid this, neat will automatically attempt to make the imported rules compliant by adding default
        values for missing information, attaching dangling properties to default containers based on the
        property type, etc.

        One has to be aware that NEAT will be opinionated about how to make the ontology
        compliant, and that the resulting rules may not be what you expect.

    """

    def _to_rules_components(
        self,
    ) -> dict:
        classes, issue_list = parse_classes(self.graph, CLASSES_QUERY, self.language, self.issue_list)
        self.issue_list = issue_list

        # NeatError
        properties, issue_list = parse_properties(self.graph, PROPERTIES_QUERY, self.language, self.issue_list)
        self.issue_list = issue_list

        components = {
            "Metadata": self._metadata,
            "Classes": list(classes.values()) if classes else [],
            "Properties": list(properties.values()) if properties else [],
        }

        return components

    @property
    def description(self) -> str:
        return f"Ontology {self.source_name} read as unverified data model"



from typing import Literal

from pydantic import BaseModel


class NeatConfig(BaseModel, validate_assignment=True):
    progress_bar: Literal["tqdm", "rich", "tqdm-notebook", "infer"] | None = "infer"
    use_iterate_bar_threshold: int | None = 500


GLOBAL_CONFIG = NeatConfig()



from collections.abc import Sequence
from typing import Literal, TypeVar, get_args

from rich.prompt import IntPrompt, Prompt

from cognite.neat._rules._constants import PATTERNS

RDFFileType = Literal["Ontology", "IMF Types", "Inference"]
NeatObjectType = Literal["Data Model", "Instances"]
XMLFileType = Literal["dexpi", "aml"]


def object_wizard(message: str = "Select object") -> NeatObjectType:
    return _selection(message, get_args(NeatObjectType))


def xml_format_wizard(message: str = "Select XML format") -> XMLFileType:
    return _selection(message, get_args(XMLFileType))


def rdf_dm_wizard(message: str = "Select source:") -> RDFFileType:
    return _selection(message, get_args(RDFFileType))


_T_Option = TypeVar("_T_Option")


def _selection(message: str, options: Sequence[_T_Option]) -> _T_Option:
    option_text = "\n  ".join([f"{i + 1}) {option}" for i, option in enumerate(options)])
    selected_index = (
        IntPrompt().ask(f"{message}\n  {option_text}\n", choices=list(map(str, range(1, len(options) + 1)))) - 1
    )
    return options[selected_index]


def space_wizard(message: str = "Set space", space: str | None = None) -> str:
    while True:
        user_input = space or Prompt().ask(f"{message}:")
        if PATTERNS.space_compliance.match(str(user_input)):
            return user_input
        else:
            print(f"Invalid input. Please provide a valid space name. {PATTERNS.space_compliance.pattern}")

        space = ""



from dataclasses import dataclass, field
from datetime import datetime
from typing import Any

import pandas as pd
from cognite.client import data_modeling as dm
from rdflib import Namespace, URIRef

from cognite.neat._constants import DEFAULT_NAMESPACE
from cognite.neat._rules.models._base_input import InputComponent, InputRules
from cognite.neat._rules.models.data_types import DataType
from cognite.neat._rules.models.entities import (
    ClassEntity,
    MultiValueTypeInfo,
    UnknownEntity,
    load_value_type,
)
from cognite.neat._utils.rdf_ import uri_display_name

from ._rules import (
    InformationClass,
    InformationMetadata,
    InformationProperty,
    InformationRules,
)


@dataclass
class InformationInputMetadata(InputComponent[InformationMetadata]):
    space: str
    external_id: str
    version: str
    creator: str
    name: str | None = None
    description: str | None = None
    created: datetime | str | None = None
    updated: datetime | str | None = None
    physical: str | URIRef | None = None
    conceptual: str | URIRef | None = None
    source_id: str | URIRef | None = None

    @classmethod
    def _get_verified_cls(cls) -> type[InformationMetadata]:
        return InformationMetadata

    def dump(self, **kwargs) -> dict[str, Any]:
        output = super().dump()
        if self.created is None:
            output["created"] = datetime.now()
        if self.updated is None:
            output["updated"] = datetime.now()
        return output

    def as_data_model_id(self) -> dm.DataModelId:
        return dm.DataModelId(space=self.space, external_id=self.external_id, version=self.version)

    @property
    def prefix(self) -> str:
        return self.space

    @property
    def identifier(self) -> URIRef:
        """Globally unique identifier for the data model.

        !!! note
            Unlike namespace, the identifier does not end with "/" or "#".

        """
        return DEFAULT_NAMESPACE[f"data-model/unverified/logical/{self.space}/{self.external_id}/{self.version}"]

    @property
    def namespace(self) -> Namespace:
        """Namespace for the data model used for the entities in the data model."""
        return Namespace(f"{self.identifier}/")


@dataclass
class InformationInputProperty(InputComponent[InformationProperty]):
    class_: ClassEntity | str
    property_: str
    value_type: DataType | ClassEntity | MultiValueTypeInfo | UnknownEntity | str
    name: str | None = None
    description: str | None = None
    min_count: int | None = None
    max_count: int | float | None = None
    default: Any | None = None
    instance_source: str | list[str] | None = None
    # Only used internally
    inherited: bool = False
    neatId: str | URIRef | None = None

    # linking
    physical: str | URIRef | None = None
    conceptual: str | URIRef | None = None

    @classmethod
    def _get_verified_cls(cls) -> type[InformationProperty]:
        return InformationProperty

    def dump(self, default_prefix: str, **kwargs) -> dict[str, Any]:  # type: ignore[override]
        output = super().dump()
        output["Class"] = ClassEntity.load(self.class_, prefix=default_prefix)
        output["Value Type"] = load_value_type(self.value_type, default_prefix)
        return output


@dataclass
class InformationInputClass(InputComponent[InformationClass]):
    class_: ClassEntity | str
    name: str | None = None
    description: str | None = None
    implements: str | list[ClassEntity] | None = None
    instance_source: str | None = None
    neatId: str | URIRef | None = None
    # linking
    physical: str | URIRef | None = None
    conceptual: str | URIRef | None = None

    @classmethod
    def _get_verified_cls(cls) -> type[InformationClass]:
        return InformationClass

    @property
    def class_str(self) -> str:
        return str(self.class_)

    def dump(self, default_prefix: str, **kwargs) -> dict[str, Any]:  # type: ignore[override]
        output = super().dump()
        parent: list[ClassEntity] | None = None
        if isinstance(self.implements, str):
            self.implements = self.implements.strip()
            parent = [ClassEntity.load(parent, prefix=default_prefix) for parent in self.implements.split(",")]
        elif isinstance(self.implements, list):
            parent = [ClassEntity.load(parent_, prefix=default_prefix) for parent_ in self.implements]
        output["Class"] = ClassEntity.load(self.class_, prefix=default_prefix)
        output["Implements"] = parent
        return output


@dataclass
class InformationInputRules(InputRules[InformationRules]):
    metadata: InformationInputMetadata
    properties: list[InformationInputProperty] = field(default_factory=list)
    classes: list[InformationInputClass] = field(default_factory=list)
    prefixes: dict[str, Namespace] | None = None

    @classmethod
    def _get_verified_cls(cls) -> type[InformationRules]:
        return InformationRules

    def dump(self) -> dict[str, Any]:
        default_prefix = self.metadata.prefix

        return dict(
            Metadata=self.metadata.dump(),
            Properties=[prop.dump(default_prefix) for prop in self.properties],
            Classes=[class_.dump(default_prefix) for class_ in self.classes],
            Prefixes=self.prefixes,
        )

    @classmethod
    def display_type_name(cls) -> str:
        return "UnverifiedInformationModel"

    @property
    def display_name(self):
        return uri_display_name(self.metadata.identifier)

    def _repr_html_(self) -> str:
        summary = {
            "type": "Logical Data Model",
            "intended for": "Information Architect",
            "name": self.metadata.name,
            "external_id": self.metadata.external_id,
            "space": self.metadata.space,
            "version": self.metadata.version,
            "classes": len(self.classes),
            "properties": len(self.properties),
        }

        return pd.DataFrame([summary]).T.rename(columns={0: ""})._repr_html_()  # type: ignore



from abc import ABC, abstractmethod
from collections.abc import Hashable, ItemsView, Iterable, Iterator, KeysView, Mapping, MutableMapping, ValuesView
from typing import (
    Any,
    Literal,
    TypeAlias,
    TypeVar,
    cast,
    final,
)

import pandas as pd
import yaml
from cognite.client.data_classes._base import T_CogniteResource
from cognite.client.data_classes.data_modeling import (
    ContainerApply,
    ContainerId,
    DataModelApply,
    DataModelId,
    NodeApply,
    NodeId,
    SpaceApply,
    ViewApply,
    ViewId,
)
from cognite.client.utils._auxiliary import load_yaml_or_json
from cognite.client.utils._pandas_helpers import convert_nullable_int_cols

Component: TypeAlias = Literal["spaces", "data_models", "views", "containers", "node_types"]

T_ID = TypeVar("T_ID", bound=Hashable)


class CogniteResourceDict(dict, MutableMapping[T_ID, T_CogniteResource], ABC):
    """CogniteResource stored in a mapping structure.

    The serialization format of the CognitiveResourceDict is a list of dicts, where each dict
    represents a CognitiveResource.

    This means that the serialization methods .dump() and .load() return a list of dicts and
    expects a list of dicts respectively.

    In addition, the init method is slightly abused compared to a regular dict by allowing the input to be a
    list of CognitiveResources.
    """

    _RESOURCE: type[T_CogniteResource]

    def __init__(
        self,
        items: Iterable[T_CogniteResource]
        | Iterable[tuple[T_ID, T_CogniteResource]]
        | Mapping[T_ID, T_CogniteResource]
        | None = None,
    ) -> None:
        if isinstance(items, Mapping):
            super().__init__(items)
        elif isinstance(items, Iterable):
            super().__init__(item if isinstance(item, tuple) else (self._as_id(item), item) for item in items)  # type: ignore[arg-type]
        else:
            super().__init__()

    @classmethod
    @abstractmethod
    def _as_id(cls, resource: T_CogniteResource) -> T_ID:
        raise NotImplementedError

    def dump(self, camel_case: bool = True) -> list[dict[str, Any]]:
        return [value.dump(camel_case) for value in self.values()]

    def dump_yaml(self) -> str:
        return yaml.dump(self.dump(camel_case=True), sort_keys=False)

    def to_pandas(self, camel_case: bool = False) -> pd.DataFrame:
        df = pd.DataFrame(self.dump(camel_case=camel_case))
        df = convert_nullable_int_cols(df)
        return df

    def _repr_html_(self) -> str:
        # Pretty print the dataframe in Jupyter
        return self.to_pandas()._repr_html_()  # type: ignore[operator]

    @classmethod
    @final
    def load(cls: "type[T_CogniteResourceDict]", resource: Iterable[dict[str, Any]] | str) -> "T_CogniteResourceDict":
        """Load a resource from a YAML/JSON string or iterable of dict."""
        if isinstance(resource, str):
            resource = load_yaml_or_json(resource)

        if isinstance(resource, Iterable):
            return cls._load(cast(Iterable, resource))

        raise TypeError(f"Resource must be json or yaml str, or iterable of dicts, not {type(resource)}")

    @classmethod
    def _load(
        cls: "type[T_CogniteResourceDict]",
        resource_list: Iterable[dict[str, Any]],
    ) -> "T_CogniteResourceDict":
        resources = (cls._RESOURCE._load(resource) for resource in resource_list)
        return cls({cls._as_id(resource): resource for resource in resources})  # type: ignore[abstract]

    # The below methods are included to make better type hints in the IDE
    def __getitem__(self, k: T_ID) -> T_CogniteResource:
        return super().__getitem__(k)

    def __setitem__(self, k: T_ID, v: T_CogniteResource) -> None:
        super().__setitem__(k, v)

    def __delitem__(self, k: T_ID) -> None:
        super().__delitem__(k)

    def __iter__(self) -> Iterator[T_ID]:
        return super().__iter__()

    def keys(self) -> KeysView[T_ID]:  # type: ignore[override]
        return super().keys()

    def values(self) -> ValuesView[T_CogniteResource]:  # type: ignore[override]
        return super().values()

    def items(self) -> ItemsView[T_ID, T_CogniteResource]:  # type: ignore[override]
        return super().items()

    def get(self, __key: T_ID, __default: Any = ...) -> T_CogniteResource:
        return super().get(__key, __default)

    def pop(self, __key: T_ID, __default: Any = ...) -> T_CogniteResource:
        return super().pop(__key, __default)

    def popitem(self) -> tuple[T_ID, T_CogniteResource]:
        return super().popitem()

    def copy(self) -> "CogniteResourceDict[T_ID, T_CogniteResource]":
        return cast(CogniteResourceDict[T_ID, T_CogniteResource], super().copy())


T_CogniteResourceDict = TypeVar("T_CogniteResourceDict", bound=CogniteResourceDict)


class ViewApplyDict(CogniteResourceDict[ViewId, ViewApply]):
    _RESOURCE = ViewApply

    @classmethod
    def _as_id(cls, resource: ViewApply) -> ViewId:
        return resource.as_id()

    @classmethod
    def from_iterable(cls, iterable: Iterable[ViewApply]) -> "ViewApplyDict":
        return cls({view.as_id(): view for view in iterable})


class SpaceApplyDict(CogniteResourceDict[str, SpaceApply]):
    _RESOURCE = SpaceApply

    @classmethod
    def _as_id(cls, resource: SpaceApply) -> str:
        return resource.space

    @classmethod
    def from_iterable(cls, iterable: Iterable[SpaceApply]) -> "SpaceApplyDict":
        return cls({space.space: space for space in iterable})


class ContainerApplyDict(CogniteResourceDict[ContainerId, ContainerApply]):
    _RESOURCE = ContainerApply

    @classmethod
    def _as_id(cls, resource: ContainerApply) -> ContainerId:
        return resource.as_id()

    @classmethod
    def from_iterable(cls, iterable: Iterable[ContainerApply]) -> "ContainerApplyDict":
        return cls({container.as_id(): container for container in iterable})


class DataModelApplyDict(CogniteResourceDict[DataModelId, DataModelApply]):
    _RESOURCE = DataModelApply

    @classmethod
    def _as_id(cls, resource: DataModelApply) -> DataModelId:
        return resource.as_id()

    @classmethod
    def from_iterable(cls, iterable: Iterable[DataModelApply]) -> "DataModelApplyDict":
        return cls({data_model.as_id(): data_model for data_model in iterable})


class NodeApplyDict(CogniteResourceDict[NodeId, NodeApply]):
    _RESOURCE = NodeApply

    @classmethod
    def _as_id(cls, resource: NodeApply) -> NodeId:
        return resource.as_id()

    @classmethod
    def from_iterable(cls, iterable: Iterable[NodeApply]) -> "NodeApplyDict":
        return cls({node.as_id(): node for node in iterable})



import warnings


class AlphaWarning(UserWarning):
    def __init__(self, feature_name: str):
        super().__init__(f"Alpha feature '{feature_name}' is subject to change without notice")

    def warn(self) -> None:
        warnings.warn(self, stacklevel=2)


class AlphaFlags:
    manual_rules_edit = AlphaWarning("enable_manual_edit")
    same_space_properties_only_export = AlphaWarning("same-space-properties-only")
    standardize_naming = AlphaWarning("standardize_naming")
    standardize_space_and_version = AlphaWarning("standardize_space_and_version")
    data_model_subsetting = AlphaWarning("data_model_subsetting")
    ontology_read = AlphaWarning("ontology_read")
    imf_read = AlphaWarning("imf_read")
    dexpi_read = AlphaWarning("dexpi_read")
    aml_read = AlphaWarning("aml_read")
    csv_read = AlphaWarning("csv_read")
    to_ontology = AlphaWarning("to_ontology")



import sys
from datetime import datetime

if sys.version_info >= (3, 11):
    from datetime import UTC
else:
    from datetime import timezone

    UTC = timezone.utc


def epoch_now_ms() -> int:
    return int((datetime.now(UTC) - datetime(1970, 1, 1, tzinfo=UTC)).total_seconds() * 1000)


def datetime_utc_now():
    return datetime.now(UTC)



import uuid
from pathlib import Path
from typing import Literal, cast

from openpyxl import Workbook
from openpyxl.cell import Cell
from openpyxl.styles import Alignment, Border, Font, NamedStyle, PatternFill, Side
from openpyxl.utils import get_column_letter
from openpyxl.worksheet.datavalidation import DataValidation

from cognite.neat._rules._constants import EntityTypes
from cognite.neat._rules.analysis import RulesAnalysis
from cognite.neat._rules.models.entities._single_value import ClassEntity
from cognite.neat._rules.models.information._rules import InformationRules

from ._base import BaseExporter


class InstanceTemplateExporter(BaseExporter[InformationRules, Workbook]):
    """
    Converts Information Rules to a templated spreadsheet meant for capturing
    instances based on class definitions in the rules.

    Args:
        no_rows: number of rows for processing, by default 1000
        auto_identifier_type: type of automatic identifier, by default "index" based, alternative is "uuid" based
        add_drop_down_list: Add drop down selection for columns that contain linking properties, by default True

    !!! note "no_rows parameter"
        no_rows should be set to the maximum expected number of instances of any of the classes.
        By default, it is set to 1000, increase it accordingly if you have more instances.

    """

    def __init__(
        self,
        no_rows: int = 1000,
        auto_identifier_type: Literal["index", "uuid"] = "index",
        add_drop_down_list: bool = True,
    ):
        self.no_rows = no_rows
        self.auto_identifier_type = auto_identifier_type
        self.add_drop_down_list = add_drop_down_list

    @property
    def description(self) -> str:
        return "Export verified information instance template to Excel."

    def export(
        self,
        rules: InformationRules,
    ):
        workbook = Workbook()

        # Remove default sheet named "Sheet"
        workbook.remove(workbook["Sheet"])

        for class_, properties in RulesAnalysis(rules).properties_by_id_by_class().items():
            workbook.create_sheet(title=class_.suffix)

            # Add header rows
            workbook[class_.suffix].append(["identifier", *list(properties.keys())])

            if self.auto_identifier_type == "uuid":
                _add_uuid_identifiers(workbook, class_.suffix, self.no_rows)
            else:
                # Default to index-based identifier
                _add_index_identifiers(workbook, class_.suffix, self.no_rows)

            for i, property_ in enumerate(properties.values()):
                if property_.type_ == EntityTypes.object_property and self.add_drop_down_list:
                    _add_drop_down_list(
                        workbook,
                        class_.suffix,
                        get_column_letter(i + 2),
                        self.no_rows,
                        cast(ClassEntity, property_.value_type).suffix,
                        "A",
                    )

        _adjust_column_width(workbook)
        _set_header_style(workbook)

        return workbook

    def export_to_file(self, rules: InformationRules, filepath: Path) -> None:
        """Exports graph capturing sheet to excel file."""
        data = self.export(rules)
        try:
            data.save(filepath)
        finally:
            data.close()
        return None


def _add_index_identifiers(workbook: Workbook, sheet: str, no_rows: int):
    """Adds index-based auto identifier to a sheet identifier column"""
    for i in range(no_rows):
        workbook[sheet][f"A{i + 2}"] = f'=IF(ISBLANK(B{i + 2}), "","{sheet}-{i + 1}")'


def _add_uuid_identifiers(workbook: Workbook, sheet: str, no_rows: int):
    """Adds UUID-based auto identifier to a sheet identifier column"""
    for i in range(no_rows):
        workbook[sheet][f"A{i + 2}"] = f'=IF(ISBLANK(B{i + 2}), "","{sheet}-{uuid.uuid4()}")'


def _add_drop_down_list(
    workbook: Workbook,
    sheet: str,
    column: str,
    no_rows: int,
    value_sheet: str,
    value_column: str,
):
    """Adds a drop down list to a column"""
    drop_down_list = DataValidation(
        type="list",
        formula1=f"={value_sheet}!{value_column}$2:{value_column}${no_rows}",
    )

    workbook[sheet].add_data_validation(drop_down_list)

    for i in range(no_rows):
        drop_down_list.add(workbook[sheet][f"{column}{i + 2}"])


def _adjust_column_width(workbook: Workbook):
    """Adjusts the column width based on the content"""
    for sheet in workbook.sheetnames:
        for cell_tuple in workbook[sheet].columns:
            # Wrong type annotation in openpyxl
            cell = cast(Cell, cell_tuple[0])  # type: ignore[index]
            if cell.value:
                adjusted_width = (len(str(cell.value)) + 5) * 1.2
                workbook[sheet].column_dimensions[cell.column_letter].width = adjusted_width


def _set_header_style(workbook: Workbook):
    """Sets the header style for all sheets in the workbook"""
    style = NamedStyle(name="header style")
    style.font = Font(bold=True, size=16)
    side = Side(style="thin", color="000000")
    style.border = Border(left=side, right=side, top=side, bottom=side)
    workbook.add_named_style(style)

    for sheet in workbook.sheetnames:
        for cell_tuple in workbook[sheet].columns:
            # Wrong type annotation in openpyxl
            cell = cast(Cell, cell_tuple[0])  # type: ignore[index]
            workbook[sheet][f"{cell.column_letter}1"].style = style
            if f"{cell.column_letter}1" == "A1":
                workbook[sheet][f"{cell.column_letter}1"].fill = PatternFill("solid", start_color="2FB5F2")
            else:
                workbook[sheet][f"{cell.column_letter}1"].fill = PatternFill("solid", start_color="FFB202")
            workbook[sheet][f"{cell.column_letter}1"].alignment = Alignment(horizontal="center", vertical="center")



import re
import sys
from functools import cached_property
from typing import Literal

if sys.version_info >= (3, 11):
    from enum import StrEnum
else:
    from backports.strenum import StrEnum


class EntityTypes(StrEnum):
    view_non_versioned = "view_non_versioned"
    subject = "subject"
    predicate = "predicate"
    object = "object"
    class_ = "class"
    parent_class = "parent_class"
    property_ = "property"
    dms_property = "dms_property"
    information_property = "information_property"
    object_property = "ObjectProperty"
    data_property = "DatatypeProperty"
    annotation_property = "AnnotationProperty"
    object_value_type = "object_value_type"
    data_value_type = "data_value_type"  # these are strings, floats, ...
    xsd_value_type = "xsd_value_type"
    dms_value_type = "dms_value_type"
    dms_node = "dms_node"
    view = "view"
    reference_entity = "reference_entity"
    container = "container"
    datamodel = "datamodel"
    undefined = "undefined"
    multi_value_type = "multi_value_type"
    asset = "asset"
    relationship = "relationship"
    edge = "edge"
    reverse = "reverse"
    unit = "unit"
    version = "version"
    prefix = "prefix"
    space = "space"


def get_reserved_words(key: Literal["class", "view", "property", "space"]) -> list[str]:
    return {
        "class": ["Class", "class"],
        "view": [
            "Query",
            "Mutation",
            "Subscription",
            "String",
            "Int32",
            "Int64",
            "Int",
            "Float32",
            "Float64",
            "Float",
            "Timestamp",
            "JSONObject",
            "Date",
            "Numeric",
            "Boolean",
            "PageInfo",
            "File",
            "Sequence",
            "TimeSeries",
        ],
        "property": [
            "property",
            "space",
            "externalId",
            "createdTime",
            "lastUpdatedTime",
            "deletedTime",
            "edge_id",
            "node_id",
            "project_id",
            "property_group",
            "seq",
            "tg_table_name",
            "extensions",
        ],
        "space": ["space", "cdf", "dms", "pg3", "shared", "system", "node", "edge"],
    }[key]


ENTITY_PATTERN = re.compile(r"^(?P<prefix>.*?):?(?P<suffix>[^(:]*)(\((?P<content>.+)\))?$")


# REGEX FOR VALIDATIONS
MORE_THAN_ONE_NONE_ALPHANUMERIC_REGEX = r"([_-]{2,})"
PREFIX_COMPLIANCE_REGEX = r"^([a-zA-Z]+)([a-zA-Z0-9]*[_-]{0,1}[a-zA-Z0-9_-]*)([a-zA-Z0-9]*)$"

SPACE_COMPLIANCE_REGEX = (
    rf"(?!^({'|'.join(get_reserved_words('space'))})$)" r"(^[a-zA-Z][a-zA-Z0-9_-]{0,41}[a-zA-Z0-9]?$)"
)


DATA_MODEL_COMPLIANCE_REGEX = r"^[a-zA-Z]([a-zA-Z0-9_]{0,253}[a-zA-Z0-9])?$"

VIEW_ID_COMPLIANCE_REGEX = (
    rf"(?!^({'|'.join(get_reserved_words('view'))})$)" r"(^[a-zA-Z][a-zA-Z0-9_]{0,253}[a-zA-Z0-9]?$)"
)
DMS_PROPERTY_ID_COMPLIANCE_REGEX = (
    rf"(?!^({'|'.join(get_reserved_words('property'))})$)" r"(^[a-zA-Z][a-zA-Z0-9_]{0,253}[a-zA-Z0-9]?$)"
)
CLASS_ID_COMPLIANCE_REGEX = rf"(?!^({'|'.join(get_reserved_words('class'))})$)" r"(^[a-zA-Z0-9._-]{0,253}[a-zA-Z0-9]?$)"

INFORMATION_PROPERTY_ID_COMPLIANCE_REGEX = r"^(\*)|(?!^(Property|property)$)(^[a-zA-Z0-9._-]{0,253}[a-zA-Z0-9]?$)"
VERSION_COMPLIANCE_REGEX = r"^[a-zA-Z0-9]([.a-zA-Z0-9_-]{0,41}[a-zA-Z0-9])?$"


# This pattern ignores commas inside brackets
SPLIT_ON_COMMA_PATTERN = re.compile(r",(?![^(]*\))")
# This pattern ignores equal signs inside brackets
SPLIT_ON_EQUAL_PATTERN = re.compile(r"=(?![^(]*\))")

# Very special Edge Entity parsing
SPLIT_ON_EDGE_ENTITY_ARGS_PATTERN = re.compile(r"(\btype\b|\bproperties\b|\bdirection\b)\s*=\s*([^,]+)")


class _Patterns:
    @cached_property
    def more_than_one_alphanumeric(self) -> re.Pattern:
        return re.compile(MORE_THAN_ONE_NONE_ALPHANUMERIC_REGEX)

    @cached_property
    def prefix_compliance(self) -> re.Pattern[str]:
        return re.compile(PREFIX_COMPLIANCE_REGEX)

    @cached_property
    def space_compliance(self) -> re.Pattern[str]:
        return re.compile(SPACE_COMPLIANCE_REGEX)

    @cached_property
    def view_id_compliance(self) -> re.Pattern[str]:
        return re.compile(VIEW_ID_COMPLIANCE_REGEX)

    @cached_property
    def dms_property_id_compliance(self) -> re.Pattern[str]:
        return re.compile(DMS_PROPERTY_ID_COMPLIANCE_REGEX)

    @cached_property
    def class_id_compliance(self) -> re.Pattern[str]:
        return re.compile(CLASS_ID_COMPLIANCE_REGEX)

    @cached_property
    def information_property_id_compliance(self) -> re.Pattern[str]:
        return re.compile(INFORMATION_PROPERTY_ID_COMPLIANCE_REGEX)

    @cached_property
    def version_compliance(self) -> re.Pattern[str]:
        return re.compile(VERSION_COMPLIANCE_REGEX)

    def entity_pattern(
        self,
        entity: EntityTypes,
    ) -> re.Pattern:
        if entity == EntityTypes.class_:
            return self.class_id_compliance

        elif entity == EntityTypes.information_property:
            return self.information_property_id_compliance

        elif entity == EntityTypes.view:
            return self.view_id_compliance

        # container regex same as view regex
        elif entity == EntityTypes.container:
            return self.view_id_compliance

        elif entity == EntityTypes.dms_property:
            return self.dms_property_id_compliance

        elif entity == EntityTypes.version:
            return self.version_compliance

        elif entity == EntityTypes.prefix:
            return self.prefix_compliance

        elif entity == EntityTypes.space:
            return self.space_compliance

        else:
            raise ValueError(f"Unsupported entity type {entity}")


PATTERNS = _Patterns()


def get_internal_properties() -> set[str]:
    return {
        "physical",
        "logical",
        "conceptual",
        "Neat ID",
    }
